import pandas as pd

from magnumapi.cadata.CableDatabase import CableDatabase
from magnumapi.geometry.blocks.RectangularBlock import RectangularBlock
from magnumapi.geometry.blocks.CosThetaBlock import AbsoluteCosThetaBlock, RelativeCosThetaBlock
from magnumapi.geometry.definitions.LayerDefinition import LayerDefinition, SlottedLayerDefinition
from magnumapi.geometry.definitions.RectangularBlockDefinition import RectangularBlockDefinition
from magnumapi.geometry.definitions.AbsoluteCosThetaBlockDefinition import AbsoluteCosThetaBlockDefinition
from magnumapi.geometry.definitions.RelativeCosThetaBlockDefinition import RelativeCosThetaBlockDefinition
from magnumapi.geometry.Geometry import Geometry, RelativeCosThetaGeometry
from magnumapi.geometry.SlottedGeometry import SlottedGeometry, SlottedRelativeCosThetaGeometry


class GeometryBuilder():
    """ Class building a geometry object with block and layer definitions. It is capable of constructing:
    - relative cos-theta geometry
    - absolute cos-theta and rectangular geometry

    """

    def __init__(self) -> None:
        self.blocks = []
        self.layer_defs = []

    def with_block_defs(self, block_defs: list, cadata: CableDatabase) -> "GeometryBuilder":
        """ Method setting block definitions for a geometry builder instance

        :param block_defs: a list of dictionaries with block definitions: absolute and relative cos-theta, absolute
        rectangular
        :param cadata: cable database instance with suplementary cable definitoins
        :return: an updated GeometryBuilder instance
        """
        self.blocks = GeometryBuilder._initialize_blocks(block_defs, cadata)
        return self

    def with_block_df(self, block_df: pd.DataFrame, cadata: CableDatabase):
        """ Method setting block definitions for a geometry builder instance

        :param block_df: a dataframe with block definitions: absolute and relative cos-theta, absolute
        rectangular
        :param cadata: cable database instance with suplementary cable definitoins
        :return: an updated GeometryBuilder instance
        """
        lst_block_defs = [row.to_dict() for _, row in block_df.iterrows()]
        self.blocks = self._initialize_blocks(lst_block_defs, cadata)
        return self

    @classmethod
    def _initialize_blocks(cls, dct_block_defs: list, cadata: CableDatabase) -> list:
        """ Private method initializing a list of blocks from a list of block definitions

        :param block_defs: a list of dictionaries with block definitions: absolute and relative cos-theta, absolute
        rectangular
        :param cadata: cable database instance with suplementary cable definitoins
        :return: a list of initialized blocks
        """
        blocks = []
        for dct_block_def in dct_block_defs:
            if dct_block_def['type'] == 2:
                for key, value in RectangularBlock.roxie_to_magnum_dct.items():
                    dct_block_def[value] = dct_block_def.pop(key)

            BlockClass, BlockDefinitionClass = cls._get_block_and_block_definition_classes(dct_block_def)
            block_def = BlockDefinitionClass(**dct_block_def)
            block = BlockClass(block_def=block_def,
                               cable_def=cadata.get_cable_definition(block_def.condname),
                               insul_def=cadata.get_insul_definition(block_def.condname),
                               strand_def=cadata.get_strand_definition(block_def.condname),
                               conductor_def=cadata.get_conductor_definition(block_def.condname))

            blocks.append(block)
        return blocks

    @staticmethod
    def _get_block_and_block_definition_classes(block_def: dict):
        """ Private method returning a tuple of a block and block definition classes to initialize a block

        :param block_def: an input block definition dictionary
        :return: a tuple of block and block definition classes
        """
        block_def_set = set(block_def.keys())

        if block_def_set == set(RelativeCosThetaBlockDefinition.__dict__['__dataclass_fields__'].keys()):
            return RelativeCosThetaBlock, RelativeCosThetaBlockDefinition
        elif block_def_set == set(AbsoluteCosThetaBlockDefinition.__dict__['__dataclass_fields__'].keys()):
            return AbsoluteCosThetaBlock, AbsoluteCosThetaBlockDefinition
        elif block_def_set == set(RectangularBlockDefinition.__dict__['__dataclass_fields__'].keys()):
            return RectangularBlock, RectangularBlockDefinition

    def with_layer_defs(self, layer_defs: list) -> "GeometryBuilder":
        """ Method initializing layer definitions

        :param layer_defs: a list of dictionaries with layer definitions
        :return: an updated GeometryBuilder instance
        """
        self.layer_defs = self._initialize_layer_defs(layer_defs)
        return self

    def with_layer_df(self, layer_df: pd.DataFrame):
        """ Method initializing layer definitions

        :param layer_defs: a list of dictionaries with layer definitions
        :return: an updated GeometryBuilder instance
        """
        lst_layer_defs = [row.to_dict() for _, row in layer_df.iterrows()]
        self.layer_defs = self._initialize_layer_defs(lst_layer_defs)
        return self

    @classmethod
    def _initialize_layer_defs(cls, dct_layer_defs: list) -> list:
        """ Private method initialiying layer definitions

        :param dct_layer_defs: a list of dictionary layer definitions
        :return: a list of LayerDefinition objects
        """
        return [LayerDefinition(**dct_layer_def) for dct_layer_def in dct_layer_defs]

    def build(self) -> Geometry:
        """ Method building a Geometry class from initialized block and layer definitions

        :return: an intialized Geometry object (either Geometry for absolute cos-theta and rectangular or
        RelativeCosThetaGeometry for a relative cos-theta geometry)
        """
        params = {'blocks': self.blocks, 'layer_defs': self.layer_defs}

        GeometryClass = self._get_geometry_class()
        return GeometryClass(**params)

    def _get_geometry_class(self):
        """ Private method returning a geometry class suitable for blocks and layer definitions. The method performs
        consistency checks prior to returning a geometry object.

        :return: a Geometry class (either Geometry for absolute cos-theta and rectangular or
        RelativeCosThetaGeometry for a relative cos-theta geometry)
        """
        self._check_initialization_consistency()

        are_all_layers_regular = all([isinstance(layer_def, LayerDefinition) for layer_def in self.layer_defs])

        if not are_all_layers_regular:
            raise AttributeError('All layers definitions have to be LayerDefinition for Geometry object.')

        are_all_blocks_relative = all([isinstance(block, RelativeCosThetaBlock) for block in self.blocks])
        are_all_blocks_absolute = all([isinstance(block, (AbsoluteCosThetaBlock, RectangularBlock))
                                       for block in self.blocks])

        if are_all_blocks_relative:
            return RelativeCosThetaGeometry
        elif are_all_blocks_absolute:
            return Geometry
        else:
            raise AttributeError('Geometry definition contains errors:'
                                 '- blocks contain a mix of relative and absolute.')

    def _check_initialization_consistency(self) -> None:
        """ Method checking initialization consistency for a GeometryBuilder. To construct a geometr both blocks and
        layer definitions need to be initialized.

        """
        if not self.blocks:
            raise AttributeError('Blocks are not initialized! Please do so using with_block_defs method.')
        if not self.layer_defs:
            raise AttributeError('Layers are not initialized! Please do so using with_layer_defs method.')


class SlottedGeometryBuilder(GeometryBuilder):
    """ A builder class for SlottedGeometry, either absolute or relative cos-theta geometry.

    """

    def __init__(self) -> None:
        super().__init__()
        self.r_aperture = None

    def with_r_aperture(self, r_aperture: float) -> "GeometryBuilder":
        """ Method setting aperture radius for a GeometryBuilder

        :param r_aperture: aperture radius in mm
        :return: an updated GeometryBuilder instance
        """
        self.r_aperture = r_aperture
        return self

    @classmethod
    def _initialize_layer_defs(cls, dct_layer_defs: list) -> list:
        """ Private method initializing layer definitions for slotted geometry

        :param dct_layer_defs: a dictionary layer definition
        :return: a list of initialized SlottedLayerDefinition objects
        """
        return [SlottedLayerDefinition(**dct_layer_def) for dct_layer_def in dct_layer_defs]

    def build(self) -> Geometry:
        """ Method building a SlottedGeometry class from initialized block and layer definitions as well as aperture
        radius

        :return: an intialized SlottedGeometry object (either SlottedGeometry for absolute cos-theta or
        SlottedRelativeCosThetaGeometry for a relative cos-theta geometry)
        """
        params = {'blocks': self.blocks, 'layer_defs': self.layer_defs, 'r_aperture': self.r_aperture}

        GeometryClass = self._get_geometry_class()
        return GeometryClass(**params)

    def _get_geometry_class(self):
        """ Private method returning a geometry class suitable for blocks and layer definitions. The method performs
        consistency checks prior to returning a geometry object.

        :return: a SlottedGeometry class (either SlottedGeometry for absolute cos-theta or
        SlottedRelativeCosThetaGeometry for a relative cos-theta geometry)
        """
        self._check_initialization_consistency()
        if self.r_aperture is None:
            raise AttributeError('Radius has to be defined for SlottedGeometry object.')

        are_all_layers_slotted = all([isinstance(layer_def, SlottedLayerDefinition) for layer_def in self.layer_defs])

        if not are_all_layers_slotted:
            raise AttributeError('All layers definitions have to be SlottedLayerDefinition for SlottedGeometry object.')

        are_all_blocks_relative = all([isinstance(block, RelativeCosThetaBlock) for block in self.blocks])
        are_all_blocks_absolute_cos_theta = all([isinstance(block, AbsoluteCosThetaBlock) for block in self.blocks])

        if are_all_blocks_relative:
            return SlottedRelativeCosThetaGeometry
        elif are_all_blocks_absolute_cos_theta:
            return SlottedGeometry
        else:
            raise AttributeError('Slotted geometry definition contains errors:'
                                 '- blocks contain a mix of rectangular and/or relative/absolute cos-theta.')



import unittest
from unittest.mock import patch

import pandas as pd

from magnumapi.geometry.GeometryChange import GeometryChange
from magnumapi.geometry.GeometryFactory import GeometryFactory
from magnumapi.geometry.GeometryPlot import GeometryPlot
from magnumapi.geometry.primitives.Line import Line
from magnumapi.cadata.CableDatabase import CableDatabase
from tests.resource_files import read_csv_as_pd, create_resources_file_path


class CosThetaGeometryTest(unittest.TestCase):
    layer_dct_defs = [
        {"no": 1, "symm": 1, "typexy": 1, "blocks": [1, 2, 3, 4]},
        {"no": 2, "symm": 1, "typexy": 1, "blocks": [5, 6, 7]},
        {"no": 3, "symm": 1, "typexy": 1, "blocks": [8, 9, 10]},
        {"no": 4, "symm": 1, "typexy": 1, "blocks": [11, 12]}]

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_bare_absolute_angles_rectangle_roxie(self, mock_show=None):
        # arrange
        blocks_dct_defs = [
            {'no': 1, 'radius': 25.0, 'alpha': 0, 'phi': 0.57294, 'nco': 4, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 2, 'radius': 25.0, 'alpha': 26, 'phi': 23.0, 'nco': 5, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 3, 'radius': 25.0, 'alpha': 47, 'phi': 50.8, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 4, 'radius': 25.0, 'alpha': 66, 'phi': 65.5, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 5, 'radius': 39.0, 'alpha': 0, 'phi': 0.36728, 'nco': 7, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 6, 'radius': 39.0, 'alpha': 35, 'phi': 26.0, 'nco': 10, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 7, 'radius': 39.0, 'alpha': 54, 'phi': 61.0, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 8, 'radius': 53.0, 'alpha': 0, 'phi': 0.27026, 'nco': 18, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 9, 'radius': 53.0, 'alpha': 30, 'phi': 31.3, 'nco': 9, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 10, 'radius': 53.0, 'alpha': 50, 'phi': 57.0, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 11, 'radius': 67.45, 'alpha': 0, 'phi': 0.21236, 'nco': 28, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 12, 'radius': 67.45, 'alpha': 30, 'phi': 37.9, 'nco': 11, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(blocks_dct_defs, self.layer_dct_defs, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/bare_absolute_angles_rectangle_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_build_geometry_with_bare_absolute_angles_rectangle_roxie_to_df(self):
        # arrange
        blocks_def = [
            {'no': 1, 'radius': 25.0, 'alpha': 0, 'phi': 0.57294, 'nco': 4, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 2, 'radius': 25.0, 'alpha': 26, 'phi': 23.0, 'nco': 5, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 3, 'radius': 25.0, 'alpha': 47, 'phi': 50.8, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 4, 'radius': 25.0, 'alpha': 66, 'phi': 65.5, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 5, 'radius': 39.0, 'alpha': 0, 'phi': 0.36728, 'nco': 7, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 6, 'radius': 39.0, 'alpha': 35, 'phi': 26.0, 'nco': 10, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 7, 'radius': 39.0, 'alpha': 54, 'phi': 61.0, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 8, 'radius': 53.0, 'alpha': 0, 'phi': 0.27026, 'nco': 18, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 9, 'radius': 53.0, 'alpha': 30, 'phi': 31.3, 'nco': 9, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 10, 'radius': 53.0, 'alpha': 50, 'phi': 57.0, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 11, 'radius': 67.45, 'alpha': 0, 'phi': 0.21236, 'nco': 28, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 12, 'radius': 67.45, 'alpha': 30, 'phi': 37.9, 'nco': 11, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(blocks_def, self.layer_dct_defs, cadata)
        geometry.build_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/bare_absolute_angles_rectangle_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        roxie_block_def = read_csv_as_pd(
            'resources/geometry/roxie/16T/reference/roxie_block_def_bare_absolute_angles_rectangle_ref.csv')
        pd.testing.assert_frame_equal(roxie_block_def, geometry.to_block_df())

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_bare_relative_angles_rectangle_roxie_to_df(self, mock_show=None):
        # arrange
        blocks_def = [
            {'no': 1, 'radius': 25.0, 'alpha_r': 0, 'phi_r': 0.57294, 'nco': 4, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 2, 'radius': 25.0, 'alpha_r': 26, 'phi_r': 10.33328, 'nco': 5, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 3, 'radius': 25.0, 'alpha_r': 21, 'phi_r': 12.72078, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 4, 'radius': 25.0, 'alpha_r': 19, 'phi_r': 8.65904, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 5, 'radius': 39.0, 'alpha_r': 0, 'phi_r': 0.36728, 'nco': 7, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 6, 'radius': 39.0, 'alpha_r': 35, 'phi_r': 12.04384, 'nco': 10, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 7, 'radius': 39.0, 'alpha_r': 19, 'phi_r': 15.69139, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 8, 'radius': 53.0, 'alpha_r': 0, 'phi_r': 0.27026, 'nco': 18, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 9, 'radius': 53.0, 'alpha_r': 30, 'phi_r': 4.62409, 'nco': 9, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 10, 'radius': 53.0, 'alpha_r': 20, 'phi_r': 12.82996, 'nco': 2, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 11, 'radius': 67.45, 'alpha_r': 0, 'phi_r': 0.21236, 'nco': 28, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
            {'no': 12, 'radius': 67.45, 'alpha_r': 30, 'phi_r': 4.76003, 'nco': 11, 'type': 1, 'current': 13500,
             'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(blocks_def, self.layer_dct_defs, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/bare_relative_angles_rectangle_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        roxie_block_def = read_csv_as_pd(
            'resources/geometry/roxie/16T/reference/roxie_block_def_bare_relative_angles_rectangle_ref.csv')
        pd.testing.assert_frame_equal(roxie_block_def, geometry.to_block_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_bare_relative_angles_rectangle_wrong_alignment(self, mock_show=None):
        # arrange
        blocks_def = [{'no': 1, 'radius': 25.0, 'alpha_r': 0, 'phi_r': 0.57294, 'nco': 7, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 2, 'radius': 25.0, 'alpha_r': 2.34375, 'phi_r': 7.921875, 'nco': 2, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 3, 'radius': 25.0, 'alpha_r': 1.09375, 'phi_r': 8.140625, 'nco': 1, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 4, 'radius': 25.0, 'alpha_r': 0.0, 'phi_r': 9.890625, 'nco': 3, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 5, 'radius': 39.0, 'alpha_r': 0, 'phi_r': 0.36728, 'nco': 10, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 6, 'radius': 39.0, 'alpha_r': 2.03125, 'phi_r': 5.515625, 'nco': 12, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 7, 'radius': 39.0, 'alpha_r': 0.9375, 'phi_r': 6.171875, 'nco': 3, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 8, 'radius': 53.0, 'alpha_r': 0, 'phi_r': 0.27026, 'nco': 16, 'type': 1,
                       'current': 13500, 'condname': '16TIL6', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 9, 'radius': 53.0, 'alpha_r': 0.0, 'phi_r': 3.328125, 'nco': 9, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 10, 'radius': 53.0, 'alpha_r': 0.15625, 'phi_r': 5.953125, 'nco': 2, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 11, 'radius': 67.45, 'alpha_r': 0, 'phi_r': 0.21236, 'nco': 30, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 12, 'radius': 67.45, 'alpha_r': 9.84375, 'phi_r': 6.71875, 'nco': 12, 'type': 1,
                       'current': 13500, 'condname': '16TOL77B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(blocks_def, self.layer_dct_defs, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_bare_absolute_angles_rectangle_csv(self, mock_show=None):
        # arrange
        root_dir = 'resources/geometry/roxie/16T/'
        block_df_path = create_resources_file_path('%sbare_absolute_block_data_rectangle.csv' % root_dir)
        layer_df_path = create_resources_file_path('%slayer_data.csv' % root_dir)
        cadata_path = create_resources_file_path('%sroxieold_2.cadata' % root_dir)
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_csv(block_df_path, layer_df_path, cadata)
        geometry.build_blocks()
        geometry.plot_blocks(xlim=(0, 80), ylim=(0, 80))

        # assert
        geometry_ref = read_csv_as_pd('%sreference/bare_absolute_angles_rectangle_ref.csv' % root_dir)
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_bare_absolute_angles_rectangle_df(self, mock_show=None):
        # arrange
        block_df = read_csv_as_pd('resources/geometry/roxie/16T/bare_absolute_block_data_rectangle.csv')
        layer_df = read_csv_as_pd('resources/geometry/roxie/16T/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_df(block_df, layer_df, cadata)
        geometry.build_blocks()
        geometry.plot_blocks(xlim=(0, 80), ylim=(0, 80))

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/bare_absolute_angles_rectangle_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_bare_relative_angles_rectangle_csv(self, mock_show=None):
        # arrange
        root_dir = 'resources/geometry/roxie/16T/'
        block_df_path = create_resources_file_path('%sbare_relative_block_data_rectangle.csv' % root_dir)
        layer_df_path = create_resources_file_path('%slayer_data.csv' % root_dir)
        cadata_path = create_resources_file_path('%sroxieold_2.cadata' % root_dir)
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_csv(block_df_path, layer_df_path, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('%sreference/bare_relative_angles_rectangle_ref.csv' % root_dir)
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_bare_relative_angles_rectangle_df(self, mock_show=None):
        # arrange
        block_df = read_csv_as_pd('resources/geometry/roxie/16T/bare_relative_block_data_rectangle.csv')
        layer_df = read_csv_as_pd('resources/geometry/roxie/16T/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_df(block_df, layer_df, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/bare_relative_angles_rectangle_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_bare_absolute_angles_rectangle_data(self, mock_show=None):
        # arrange
        data_path = create_resources_file_path('resources/geometry/roxie/16T/16T_22b-37-optd7f8_gx.data')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_data(data_path, cadata)
        geometry.build_blocks()
        geometry.plot_blocks(xlim=(0, 80), ylim=(0, 80))

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_bare_absolute_angles_trapezoid(self, mock_show=None):
        # arrange
        block_df = read_csv_as_pd('resources/geometry/roxie/16T/bare_absolute_block_data_trapezoid.csv')
        layer_df = read_csv_as_pd('resources/geometry/roxie/16T/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_df(block_df, layer_df, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/bare_absolute_angles_trapezoid_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_bare_relative_angles_trapezoid(self, mock_show=None):
        # arrangee
        block_df = read_csv_as_pd('resources/geometry/roxie/16T/bare_relative_block_data_trapezoid.csv')
        layer_df = read_csv_as_pd('resources/geometry/roxie/16T/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_df(block_df, layer_df, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/bare_relative_angles_trapezoid_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_ins_relative_angles_rectangle(self, mock_show=None):
        # arrange
        block_df = read_csv_as_pd('resources/geometry/roxie/16T/ins_relative_block_data_rectangle.csv')
        layer_df = read_csv_as_pd('resources/geometry/roxie/16T/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_df(block_df, layer_df, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/ins_relative_angles_rectangle_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_ins_absolute_angles_trapezoid(self, mock_show=None):
        # arrange
        block_df = read_csv_as_pd('resources/geometry/roxie/16T/ins_absolute_block_data_trapezoid.csv')
        layer_df = read_csv_as_pd('resources/geometry/roxie/16T/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_df(block_df, layer_df, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/ins_absolute_angles_trapezoid_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_build_geometry_with_roxie_ins_relative_angles_trapezoid(self, mock_show=None):
        # arrange
        block_df = read_csv_as_pd('resources/geometry/roxie/16T/ins_relative_block_data_trapezoid.csv')
        layer_df = read_csv_as_pd('resources/geometry/roxie/16T/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_df(block_df, layer_df, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()
        geometry.plot_bare_blocks()

        # assert
        geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/ins_relative_angles_trapezoid_ref.csv')
        pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_check_geometry_ansys(self, mock_show=None):
        # arrange
        block_df = read_csv_as_pd('resources/geometry/roxie/16T/ins_relative_block_data_trapezoid.csv')
        layer_df = read_csv_as_pd('resources/geometry/roxie/16T/layer_data.csv')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_df(block_df, layer_df, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()
        geometry.plot_bare_blocks()

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     1     4           25            5           10        13500     16TIL9   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current    condname  n1 n2  imag         turn
    #  INSUL 1
    #   1 16TINS            0.15         0.15 'EUROCIRCOL                    '
    #  No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #   1 16TILG6          21.84       1.8782       2.2595    38          100            3 'EUROCIRCOL 26 STR             '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    #   1 16TIL6        1 16TILG6     STR11     NB3SN2    16TINS    NONE      NB3SN                1.9 'eurocircol                    '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans     QuenchMat.           T_o  Comment

    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_three_trapz_with_insulation_with_radius_with_phi_with_alpha_abs(self,
                                                                                                           mock_show=None):

        # arrange
        blocks_def = [
            {'no': 1, 'radius': 25, 'alpha': 10, 'phi': 5, 'nco': 3, 'type': 1, 'current': 13500,
             'condname': '16TOL78B',
             'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]
        layer_def = [{"no": 1, "symm": 1, "typexy": 1, "blocks": [1]}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(blocks_def, layer_def, cadata)

        geometry.build_blocks()
        geometry.plot_blocks()
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.6208
        # y(1) = 4.3379
        # x(2) = 25.0173
        # y(2) = 2.1987
        # x(3) = 46.8217
        # y(3) = 6.0434
        # x(4) = 46.3548
        # y(4) = 8.5627
        self.assertAlmostEqual(24.6208, round(geometry.blocks[0].areas[0].get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.3379, round(geometry.blocks[0].areas[0].get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.0173, round(geometry.blocks[0].areas[0].get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1987, round(geometry.blocks[0].areas[0].get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.8217, round(geometry.blocks[0].areas[0].get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(6.0434, round(geometry.blocks[0].areas[0].get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.3548, round(geometry.blocks[0].areas[0].get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(8.5627, round(geometry.blocks[0].areas[0].get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0(1 - 4) = 9.9922
        # Alph0(1 - 4) = 11.0003
        # Phi0(2 - 3) = 5.0000
        # Alph0(2 - 3) = 10.0000
        self.assertAlmostEqual(9.9922, Line.calculate_positioning_angle(geometry.blocks[0].areas[0].get_line(2),
                                                                        geometry.blocks[0].block_def.radius), places=4)
        self.assertAlmostEqual(11.0003, Line.calculate_relative_alpha_angle(geometry.blocks[0].areas[0].get_line(2)),
                               places=4)
        self.assertAlmostEqual(5.0000, Line.calculate_positioning_angle(geometry.blocks[0].areas[0].get_line(0),
                                                                        geometry.blocks[0].block_def.radius), places=4)
        self.assertAlmostEqual(10.0000, Line.calculate_relative_alpha_angle(geometry.blocks[0].areas[0].get_line(0)),
                               places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.7954
        # y(1) = 4.2190
        # x(2) = 25.1376
        # y(2) = 2.3723
        # x(3) = 46.6467
        # y(3) = 6.1649
        # x(4) = 46.2349
        # y(4) = 8.3865

        bare_blocks = geometry.get_bare_areas_for_blocks()

        self.assertAlmostEqual(24.7954, round(bare_blocks[0][0].get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.2190, round(bare_blocks[0][0].get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.1376, round(bare_blocks[0][0].get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.3723, round(bare_blocks[0][0].get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.6467, round(bare_blocks[0][0].get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(6.1649, round(bare_blocks[0][0].get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.2349, round(bare_blocks[0][0].get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(8.3865, round(bare_blocks[0][0].get_line(2).p1.y, 4), places=4)

        # Conductor number: 3
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 23.6795
        # y(1) = 8.5859
        # x(2) = 24.1505
        # y(2) = 6.4619
        # x(3) = 45.8074
        # y(3) = 11.0654
        # x(4) = 45.2528
        # y(4) = 13.5668
        self.assertAlmostEqual(23.6795, round(geometry.blocks[0].areas[-1].get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(8.5859, round(geometry.blocks[0].areas[-1].get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(24.1505, round(geometry.blocks[0].areas[-1].get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(6.4619, round(geometry.blocks[0].areas[-1].get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(45.8074, round(geometry.blocks[0].areas[-1].get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(11.0654, round(geometry.blocks[0].areas[-1].get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(45.2528, round(geometry.blocks[0].areas[-1].get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(13.5668, round(geometry.blocks[0].areas[-1].get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0(1 - 4) = 19.9823
        # Alph0(1 - 4) = 13.0009
        # Phi0(2 - 3) = 14.9796
        # Alph0(2 - 3) = 12.0006
        self.assertAlmostEqual(19.9823, Line.calculate_positioning_angle(geometry.blocks[0].areas[-1].get_line(2),
                                                                         geometry.blocks[0].block_def.radius), places=4)
        self.assertAlmostEqual(13.0009, Line.calculate_relative_alpha_angle(geometry.blocks[0].areas[-1].get_line(2)),
                               places=4)
        self.assertAlmostEqual(14.9796, Line.calculate_positioning_angle(geometry.blocks[0].areas[-1].get_line(0),
                                                                         geometry.blocks[0].block_def.radius), places=4)
        self.assertAlmostEqual(12.0006, Line.calculate_relative_alpha_angle(geometry.blocks[0].areas[-1].get_line(0)),
                               places=4)

        # assert positions - bare
        # Conductor number: 3
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 23.8582
        # y(1) = 8.4732
        # x(2) = 24.2647
        # y(2) = 6.6395
        # x(3) = 45.6282
        # y(3) = 11.1807
        # x(4) = 45.1392
        # y(4) = 13.3866

        bare_blocks = geometry.get_bare_areas_for_blocks()

        self.assertAlmostEqual(23.8582, round(bare_blocks[0][-1].get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(8.4732, round(bare_blocks[0][-1].get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(24.2647, round(bare_blocks[0][-1].get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(6.6395, round(bare_blocks[0][-1].get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(45.6282, round(bare_blocks[0][-1].get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(11.1807, round(bare_blocks[0][-1].get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(45.1392, round(bare_blocks[0][-1].get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(13.3866, round(bare_blocks[0][-1].get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_get_first_isosceles_trapezium_three_trapz_with_insulation_with_radius_with_phi_with_alpha_rel(self,
                                                                                                           mock_show=None):

        # arrange
        blocks_def = [
            {'no': 1, 'radius': 25, 'alpha_r': 10, 'phi_r': 5, 'nco': 3, 'type': 1, 'current': 13500,
             'condname': '16TOL78B', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]
        layer_def = [{"no": 1, "symm": 1, "typexy": 1, "blocks": [1]}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(blocks_def, layer_def, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.6208
        # y(1) = 4.3379
        # x(2) = 25.0173
        # y(2) = 2.1987
        # x(3) = 46.8217
        # y(3) = 6.0434
        # x(4) = 46.3548
        # y(4) = 8.5627
        self.assertAlmostEqual(24.6208, round(geometry.blocks[0].areas[0].get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.3379, round(geometry.blocks[0].areas[0].get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.0173, round(geometry.blocks[0].areas[0].get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.1987, round(geometry.blocks[0].areas[0].get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.8217, round(geometry.blocks[0].areas[0].get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(6.0434, round(geometry.blocks[0].areas[0].get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.3548, round(geometry.blocks[0].areas[0].get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(8.5627, round(geometry.blocks[0].areas[0].get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0(1 - 4) = 9.9922
        # Alph0(1 - 4) = 11.0003
        # Phi0(2 - 3) = 5.0000
        # Alph0(2 - 3) = 10.0000
        self.assertAlmostEqual(9.9922, Line.calculate_positioning_angle(geometry.blocks[0].areas[0].get_line(2),
                                                                        geometry.blocks[0].block_def.radius), places=4)
        self.assertAlmostEqual(11.0003, Line.calculate_relative_alpha_angle(geometry.blocks[0].areas[0].get_line(2)),
                               places=4)
        self.assertAlmostEqual(5.0000, Line.calculate_positioning_angle(geometry.blocks[0].areas[0].get_line(0),
                                                                        geometry.blocks[0].block_def.radius), places=4)
        self.assertAlmostEqual(10.0000, Line.calculate_relative_alpha_angle(geometry.blocks[0].areas[0].get_line(0)),
                               places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 24.7954
        # y(1) = 4.2190
        # x(2) = 25.1376
        # y(2) = 2.3723
        # x(3) = 46.6467
        # y(3) = 6.1649
        # x(4) = 46.2349
        # y(4) = 8.3865

        bare_blocks = geometry.get_bare_areas_for_blocks()

        self.assertAlmostEqual(24.7954, round(bare_blocks[0][0].get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(4.2190, round(bare_blocks[0][0].get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(25.1376, round(bare_blocks[0][0].get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(2.3723, round(bare_blocks[0][0].get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(46.6467, round(bare_blocks[0][0].get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(6.1649, round(bare_blocks[0][0].get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(46.2349, round(bare_blocks[0][0].get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(8.3865, round(bare_blocks[0][0].get_line(2).p1.y, 4), places=4)

        # Conductor number: 3
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 23.6795
        # y(1) = 8.5859
        # x(2) = 24.1505
        # y(2) = 6.4619
        # x(3) = 45.8074
        # y(3) = 11.0654
        # x(4) = 45.2528
        # y(4) = 13.5668
        self.assertAlmostEqual(23.6795, round(geometry.blocks[0].areas[-1].get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(8.5859, round(geometry.blocks[0].areas[-1].get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(24.1505, round(geometry.blocks[0].areas[-1].get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(6.4619, round(geometry.blocks[0].areas[-1].get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(45.8074, round(geometry.blocks[0].areas[-1].get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(11.0654, round(geometry.blocks[0].areas[-1].get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(45.2528, round(geometry.blocks[0].areas[-1].get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(13.5668, round(geometry.blocks[0].areas[-1].get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0(1 - 4) = 19.9823
        # Alph0(1 - 4) = 13.0009
        # Phi0(2 - 3) = 14.9796
        # Alph0(2 - 3) = 12.0006
        self.assertAlmostEqual(19.9823, Line.calculate_positioning_angle(geometry.blocks[0].areas[-1].get_line(2),
                                                                         geometry.blocks[0].block_def.radius), places=4)
        self.assertAlmostEqual(13.0009, Line.calculate_relative_alpha_angle(geometry.blocks[0].areas[-1].get_line(2)),
                               places=4)
        self.assertAlmostEqual(14.9796, Line.calculate_positioning_angle(geometry.blocks[0].areas[-1].get_line(0),
                                                                         geometry.blocks[0].block_def.radius), places=4)
        self.assertAlmostEqual(12.0006, Line.calculate_relative_alpha_angle(geometry.blocks[0].areas[-1].get_line(0)),
                               places=4)

        # assert positions - bare
        # Conductor number: 3
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 23.8582
        # y(1) = 8.4732
        # x(2) = 24.2647
        # y(2) = 6.6395
        # x(3) = 45.6282
        # y(3) = 11.1807
        # x(4) = 45.1392
        # y(4) = 13.3866

        bare_blocks = geometry.get_bare_areas_for_blocks()

        self.assertAlmostEqual(23.8582, round(bare_blocks[0][-1].get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(8.4732, round(bare_blocks[0][-1].get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(24.2647, round(bare_blocks[0][-1].get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(6.6395, round(bare_blocks[0][-1].get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(45.6282, round(bare_blocks[0][-1].get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(11.1807, round(bare_blocks[0][-1].get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(45.1392, round(bare_blocks[0][-1].get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(13.3866, round(bare_blocks[0][-1].get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_wrongly_generated_geometry(self, mock_show=None):

        blocks_def = [{'no': 1, 'radius': 25.0, 'alpha_r': 0, 'phi_r': 0.57294, 'nco': 3, 'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 2, 'radius': 25.0, 'alpha_r': 2.1875, 'phi_r': 7.375, 'nco': 4, 'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 3, 'radius': 25.0, 'alpha_r': 9.84375, 'phi_r': 4.09375, 'nco': 2, 'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 4, 'radius': 25.0, 'alpha_r': 6.71875, 'phi_r': 6.171875, 'nco': 2, 'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 5, 'radius': 39.0, 'alpha_r': 0.0, 'phi_r': 0.36728, 'nco': 3, 'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 6, 'radius': 39.0, 'alpha_r': 8.59375, 'phi_r': 8.25, 'nco': 9, 'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 7, 'radius': 39.0, 'alpha_r': 9.84375, 'phi_r': 3.21875, 'nco': 3, 'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 8, 'radius': 53.0, 'alpha_r': 0, 'phi_r': 0.27026, 'nco': 18, 'type': 1,
                       'current': 13500, 'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 9, 'radius': 53.0, 'alpha_r': 9.84375, 'phi_r': 9.5625, 'nco': 12, 'type': 1,
                       'current': 13500, 'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 10, 'radius': 53.0, 'alpha_r': 3.28125, 'phi_r': 6.828125, 'nco': 0, 'type': 1,
                       'current': 13500, 'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 11, 'radius': 67.45, 'alpha_r': 0, 'phi_r': 0.21236, 'nco': 30, 'type': 1,
                       'current': 13500, 'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 12, 'radius': 67.45, 'alpha_r': 0.15625, 'phi_r': 3.0, 'nco': 12, 'type': 1,
                       'current': 13500, 'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(blocks_def, self.layer_dct_defs, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("matplotlib.pyplot.show")
    def test_simplified_magnet_input(self, mock_show=None):
        # arrange
        blocks_def = [{'no': 1, 'radius': 25.0, 'alpha_r': 0, 'phi_r': 0.57294, 'nco': 4, 'type': 1, 'current': 13500,
                       'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 2, 'radius': 25.0, 'alpha_r': 23.9998691192517, 'phi_r': 1.9732107800086354, 'nco': 5,
                       'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 3, 'radius': 25.0, 'alpha_r': 18.499836399064584, 'phi_r': 2.2613872613106736, 'nco': 2,
                       'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 4, 'radius': 25.0, 'alpha_r': 17.999934559625828, 'phi_r': 4.545439476728973, 'nco': 2,
                       'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 5, 'radius': 39.0, 'alpha_r': 0.0, 'phi_r': 0.36728, 'nco': 7, 'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 6, 'radius': 39.0, 'alpha_r': 31.499770958690473, 'phi_r': 2.6566334521150843, 'nco': 10,
                       'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 7, 'radius': 39.0, 'alpha_r': 13.99967279812931, 'phi_r': 2.4079874063247146, 'nco': 2,
                       'type': 1,
                       'current': 13500, 'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 8, 'radius': 53.0, 'alpha_r': 0, 'phi_r': 0.27026, 'nco': 18, 'type': 1, 'current': 13500,
                       'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 9, 'radius': 53.0, 'alpha_r': 21.00147158488118, 'phi_r': 1.1333184204627678, 'nco': 9,
                       'type': 1,
                       'current': 13500, 'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 10, 'radius': 53.0, 'alpha_r': 15.500735792440558, 'phi_r': 10.975148087526826, 'nco': 2,
                       'type': 1,
                       'current': 13500, 'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 11, 'radius': 67.45, 'alpha_r': 0, 'phi_r': 0.21236, 'nco': 28, 'type': 1,
                       'current': 13500,
                       'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0},
                      {'no': 12, 'radius': 67.45, 'alpha_r': 16.002289132037205, 'phi_r': 1.0131471917431227, 'nco': 11,
                       'type': 1,
                       'current': 13500, 'condname': '16TOL8', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(blocks_def, self.layer_dct_defs, cadata)
        geometry.build_blocks()
        geometry.plot_blocks()

        # assert
        # ToDo - fix this test
        # geometry_ref = read_csv_as_pd('resources/geometry/roxie/16T/reference/bare_absolute_angles_rectangle_ref.csv')
        # pd.testing.assert_frame_equal(geometry_ref, geometry.to_df())

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("plotly.graph_objects.Figure.show")
    def test_simplified_magnet_input_plotly_blocks(self, mock_show=None):
        # arrange
        block_def = [{'no': 1, 'radius': 25.0, 'alpha_r': 0, 'phi_r': 0.57294, 'nco': 4, 'type': 1, 'current': 13500,
                      'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]
        layer_def = [{"no": 1, "symm": 1, "typexy": 1, "blocks": [1]}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(block_def, layer_def, cadata)
        geometry.build_blocks()
        geometry.plotly_blocks()

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    @patch("plotly.graph_objects.Figure.show")
    def test_simplified_magnet_input_plotly_homogenized_blocks(self, mock_show=None):
        # arrange
        block_def = [{'no': 1, 'radius': 25.0, 'alpha_r': 0, 'phi_r': 0.57294, 'nco': 4, 'type': 1, 'current': 13500,
                      'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]
        layer_def = [{"no": 1, "symm": 1, "typexy": 1, "blocks": [1]}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(block_def, layer_def, cadata)
        geometry.build_blocks()

        homo_geometry = geometry.homogenize()
        homo_geometry.build_blocks()
        homo_geometry.plotly_blocks()

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_simplified_magnet_input_display_table(self):
        # arrange
        block_def = [{'no': 1, 'radius': 25.0, 'alpha_r': 0, 'phi_r': 0.57294, 'nco': 4, 'type': 1, 'current': 13500,
                      'condname': '16TIL9', 'n1': 2, 'n2': 20, 'imag': 0, 'turn': 0}]
        layer_def = [{"no": 1, "symm": 1, "typexy": 1, "blocks": [1]}]

        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_dict(block_def, layer_def, cadata)
        geometry.build_blocks()
        GeometryPlot.display_definition_table(geometry.to_df())

    def test_with_roxie_absolute_angle_definition_json(self):
        # arrange
        json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_abs.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_json(json_path, cadata)

        current = geometry.blocks[0].block_def.current
        f_cu_nocu = geometry.blocks[0].strand_def.f_cu_nocu
        temp_ref = geometry.blocks[0].conductor_def.temp_ref
        rrr = geometry.blocks[0].strand_def.rrr
        d_strand = geometry.blocks[0].strand_def.d_strand
        n_s = geometry.blocks[0].cable_def.n_s

        # assert
        self.assertEqual(13500, current)
        self.assertAlmostEqual(0.85, f_cu_nocu, places=2)
        self.assertAlmostEqual(1.9, temp_ref, places=2)
        self.assertEqual(100, rrr)
        self.assertAlmostEqual(1.1, d_strand, places=2)
        self.assertEqual(22, n_s)

    def test_with_roxie_relative_angle_definition_json(self):
        # arrange
        json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_json(json_path, cadata)

        current = geometry.blocks[0].block_def.current
        f_cu_nocu = geometry.blocks[0].strand_def.f_cu_nocu
        temp_ref = geometry.blocks[0].conductor_def.temp_ref
        rrr = geometry.blocks[0].strand_def.rrr
        d_strand = geometry.blocks[0].strand_def.d_strand
        n_s = geometry.blocks[0].cable_def.n_s

        # assert
        self.assertEqual(13500, current)
        self.assertAlmostEqual(0.85, f_cu_nocu, places=2)
        self.assertAlmostEqual(1.9, temp_ref, places=2)
        self.assertEqual(100, rrr)
        self.assertAlmostEqual(1.1, d_strand, places=2)
        self.assertEqual(22, n_s)

    @patch("plotly.graph_objects.Figure.show")
    def test_with_roxie_absolute_angle_definition_json_radiality_corection(self, mock_show=None):
        # arrange
        json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_abs.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)

        # act
        geometry = GeometryFactory.init_with_json(json_path, cadata)

        geometry.build_blocks()
        geometry.plotly_blocks(figsize=(90, 90), xlim=(0, 85))

        geometry = GeometryChange.calculate_radial_alpha(geometry)

        # Build corrected blocks
        geometry.build_blocks()

        # Display corrected blocks
        geometry.plotly_blocks(figsize=(90, 90), xlim=(0, 85))

        # assert
        if mock_show is not None:
            mock_show.assert_called()



from dataclasses import dataclass

from magnumapi.cadata.Definition import Definition


@dataclass
class QuenchDefinition(Definition):
    """Class for quench definition.

       Attributes:
           cp_sc (str): The fit function for the specific heat of the superconductor.
           See also: http://cern.ch/roxie > Documentation > Materials.pdf
           cp_cu (str): The fit function for the specific heat of copper.
           See also: http://cern.ch/roxie > Documentation > Materials.pdf
           k_cu (str): The fit function for the thermal conductivity of copper.
           See also: http://cern.ch/roxie > Documentation > Materials.pdf
           res_cu (str): The fit function for the electrical resistitvity of copper.
           See also: http://cern.ch/roxie > Documentation > Materials.pdf
           cp_ins (str): The fit function for the heat capacity of the cable insulation.
           See also: http://cern.ch/roxie > Documentation > Materials.pdf
           k_ins (str): The fit function for the thermal conductivity of the cable insulation.
           See also: http://cern.ch/roxie > Documentation > Materials.pdf
           cp_fill (str): The fit function for the heat capacity of the material filling the cable voids.
           See also: http://cern.ch/roxie > Documentation > Materials.pdf
           perc_he (float): The percentage of the cable voids that is filled by helium.

    """
    cp_sc: str = ''
    cp_cu: str = ''
    k_cu: str = ''
    res_cu: str = ''
    cp_ins: str = ''
    k_ins: str = ''
    cp_fill: str = ''
    perc_he: float = ''

    @staticmethod
    def get_magnum_to_roxie_dct() -> dict:
        return {"name": "Name",
                "cp_sc": "SCHeatCapa",
                "cp_cu": "CuHeatCapa",
                "k_cu": "CuThermCond",
                "res_cu": "CuElecRes",
                "cp_ins": "InsHeatCapa",
                "k_ins": "InsThermCond",
                "cp_fill": "FillHeatCapa",
                "perc_he": "He%",
                "comment": "Comment"}



from unittest import TestCase
from unittest.mock import patch

import pandas as pd
import numpy as np

from magnumapi.optimization.GeneticOptimization import GeneticOptimization
from magnumapi.optimization.constants import PENALTY
from tests.resource_files import create_resources_file_path

pop = [[1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1],
       [0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0],
       [1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0],
       [0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1],
       [1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1],
       [0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1],
       [1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0],
       [1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1],
       [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
       [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]

scores = [1, 1.2, 2, 2.5, 3, 3, 4, 5, 10, 20]

json_file_path = create_resources_file_path('resources/optimization/config.json')
optimization_cfg = GeneticOptimization.initialize_config(json_file_path)
csv_path = create_resources_file_path('resources/optimization/GeneticOptimization.csv')
logger_df = pd.read_csv(csv_path, index_col=0)
logger_lst = [pd.DataFrame(record, index=[0]) for record in logger_df.to_dict('records')]


class TestGeneticOptimization(TestCase):

    @patch.multiple(GeneticOptimization, __abstractmethods__=set())
    def setUp(self) -> None:
        csv_file_path = create_resources_file_path('resources/optimization/optim_input_enlarged.csv')
        self.gen_opt = GeneticOptimization(optimization_cfg, pd.read_csv(csv_file_path))
        self.gen_opt.logger.logs = logger_lst

    def test_str_config(self):
        # arrange

        # act

        # assert
        str_repr_ref_path = create_resources_file_path('resources/optimization/config_str_representation_ref.txt')

        with open(str_repr_ref_path, 'r') as file:
            str_repr_ref = file.read()

        self.assertEqual(str_repr_ref, str(optimization_cfg))

    def test_get_logger_df(self):
        # arrange

        # act
        logger_df_act = self.gen_opt.get_logger_df()

        # assert
        pd.testing.assert_frame_equal(logger_df, logger_df_act)

    def test_get_mean_fitness_per_generation(self):
        # arrange

        # act
        logger_mean_df_act = self.gen_opt.get_mean_fitness_per_generation()

        # assert
        logger_mean_df_csv_path = create_resources_file_path('resources/optimization/GeneticOptimizationMean.csv')
        logger_mean_df = pd.read_csv(logger_mean_df_csv_path, index_col=0)
        pd.testing.assert_frame_equal(logger_mean_df, logger_mean_df_act)

    def test_get_min_fitness_per_generation(self):
        # arrange

        # act
        logger_min_df_act = self.gen_opt.get_min_fitness_per_generation()

        # assert
        logger_min_df_csv_path = create_resources_file_path('resources/optimization/GeneticOptimizationMin.csv')
        logger_mean_df = pd.read_csv(logger_min_df_csv_path, index_col=0)
        pd.testing.assert_frame_equal(logger_mean_df, logger_min_df_act)

    def test_create_new_generation_without_elitism(self):
        # arrange
        np.random.seed(0)

        # act
        self.gen_opt.n_elite = 0
        pop_act = self.gen_opt.update_generation(pop, scores)

        # assert
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[0])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1], pop_act[1])
        self.assertListEqual([1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1], pop_act[2])
        self.assertListEqual([0, 1, 0, 1, 1, 1, 0, 0, 1, 1, 0], pop_act[3])
        self.assertListEqual([0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0], pop_act[4])
        self.assertListEqual([1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0], pop_act[5])
        self.assertListEqual([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], pop_act[6])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[7])
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[8])
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[9])

    def test_create_new_generation_with_elitism(self):
        # arrange
        np.random.seed(0)

        # act
        self.gen_opt.n_elite = 2
        pop_act = self.gen_opt.update_generation(pop, scores)

        # assert
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[0])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[1])
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[2])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1], pop_act[3])
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1], pop_act[4])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[5])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0], pop_act[6])
        self.assertListEqual([1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0], pop_act[7])
        self.assertListEqual([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], pop_act[8])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[9])

    def test_selection(self):
        # arrange
        np.random.seed(0)

        # act
        individual = GeneticOptimization.selection(pop, scores)

        # assert
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], individual)

    def test_crossover(self):
        # arrange
        np.random.seed(0)

        p1 = [1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1]
        p2 = [0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0]

        # act
        c = self.gen_opt.crossover(p1, p2)

        # assert
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0], c[0])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1], c[1])

    def test_mutation(self):
        # arrange
        np.random.seed(0)
        chromosome = 10 * [1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1]

        # act
        mutated_chromosome = self.gen_opt.mutation(chromosome)

        # assert
        mutated_chromosome_ref = [1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0,
                                  1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0,
                                  1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0,
                                  1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1]

        self.assertEqual(mutated_chromosome_ref, mutated_chromosome)

    def test_calculate_score_with_nan(self):
        # arrange
        fom_dct = {'b3': float('nan'), 'b5': float('nan'), 'margmi': float('nan'), 'seqv': float('nan')}

        # act
        score = self.gen_opt.calculate_score(fom_dct)

        # assert
        self.assertEqual(PENALTY, score)

    def test_calculate_score(self):
        # arrange
        fom_dct = {'b3': 10.5, 'b5': 200, 'margmi': 10, 'seqv': 111}

        # act
        score = self.gen_opt.calculate_score(fom_dct)

        # assert
        self.assertAlmostEqual(30.311, score, places=5)



from dataclasses import dataclass
from typing import List
import magnumapi.commons.json_file as json_file

from magnumapi.optimization.ObjectiveConfig import ObjectiveConfig
from magnumapi.optimization.OptimizationNotebookConfig import OptimizationNotebookConfig


@dataclass
class OptimizationConfig:
    """Class for optimization config used for the genetic algorithm.

    Attributes:
       input_folder (str): The path to an input folder
       output_folder (float): The path to an output folder
       logger_rel_path (float): The path to the logger csv file
       n_pop (int): The number of individuals, i.e., the population size
       n_gen (int): The number of generations
       r_cross (float): The probability of crossover
       r_mut (float): The probability of mutation
       objectives (list): The list of objective configs
       notebooks (list): The list of notebook configs
       model_creation_type (str): type of model creation: default, programmable, targeted
    """
    input_folder: str
    output_folder: str
    logger_rel_path: str
    n_pop: int
    n_gen: int
    r_cross: float
    r_mut: float
    objectives: List[ObjectiveConfig]
    notebooks: List[OptimizationNotebookConfig]
    model_creation_type: str

    def __str__(self) -> str:
        notebooks_str = "\n\n".join(str(notebook) for notebook in self.notebooks)
        objectives_str = "\n\n".join(str(objective) for objective in self.objectives)
        return "input_folder: %s\n" \
               "output_folder: %s\n" \
               "logger_rel_path: %s\n" \
               "n_pop: %d\n" \
               "n_gen: %d\n" \
               "r_cross: %f\n" \
               "r_mut: %f\n" \
               "objectives: \n\n" \
               "%snotebooks: \n\n" \
               "%s" % (self.input_folder,
                       self.output_folder,
                       self.logger_rel_path,
                       self.n_pop,
                       self.n_gen,
                       self.r_cross,
                       self.r_mut,
                       objectives_str,
                       notebooks_str)

    @staticmethod
    def initialize_config(json_path: str) -> "OptimizationConfig":
        """ Static method initializing an optimization config from a json file

        :param json_path: a path to a json file with config
        :return: initialized OptimizationConfig instance
        """
        data = json_file.read(json_path)

        input_folder = data['input_folder']
        output_folder = data['output_folder']
        logger_rel_path = data['logger_rel_path']
        n_pop = data['n_pop']
        n_gen = data['n_gen']
        r_cross = data['r_cross']
        r_mut = data['r_mut']
        model_creation_type = data.get('model_creation_type', 'programmable')
        objectives = [ObjectiveConfig(**ff) for ff in data['objectives']]
        notebooks = [OptimizationNotebookConfig(**nb) for nb in data['notebooks']]

        return OptimizationConfig(input_folder=input_folder,
                                  output_folder=output_folder,
                                  logger_rel_path=logger_rel_path,
                                  n_pop=n_pop,
                                  n_gen=n_gen,
                                  r_cross=r_cross,
                                  r_mut=r_mut,
                                  objectives=objectives,
                                  notebooks=notebooks,
                                  model_creation_type=model_creation_type)

    def get_weight(self, objective: str) -> float:
        """ Method iterating through the list of objective config, finding a matching name and returning
        corresponding weight.

        :param objective: name of an objective variable
        :return: value of the objective weight
        """
        for config_objective in self.objectives:
            if config_objective.objective == objective:
                return config_objective.weight

        raise KeyError('Objective name %s not present in objective configs.' % objective)

    def get_constraint(self, objective: str) -> float:
        """ Method iterating through the list of objective config, finding a matching name and returning
        corresponding constraint.

        :param objective: name of an objective variable
        :return: value of the objective constraint
        """
        for config_objective in self.objectives:
            if config_objective.objective == objective:
                return config_objective.constraint

        raise KeyError('Objective name %s not present in objective configs.' % objective)



import os
from typing import Dict, List

import pandas as pd

from magnumapi.optimization.constants import SCORE_KEYWORD, PENALTY


class Logger:
    """ Class for logging of genetic optimization progress.

    """
    def __init__(self, output_subdirectory_dir: str, logger_rel_path: str) -> None:
        """ A constructor of a Logger instance

        :param output_subdirectory_dir: a root subdirectory with output results
        """
        self.logs: List[pd.DataFrame, ...] = []
        self.output_subdirectory_dir = output_subdirectory_dir
        self.logger_rel_path = logger_rel_path

    def append_to_logger(self, decoded_chromosome: Dict, fom_dct: Dict, score: float) -> None:
        """ Abstract method updating model parameters

        :param decoded_chromosome: a dictionary with a decoded chromosome
        :param fom_dct: a dictionary with figures of merit
        :param score: score of an individual
        """
        self.logs.append(pd.DataFrame({**decoded_chromosome, **fom_dct, **{SCORE_KEYWORD: score}}, index=[0]))

    def get_logger_df(self) -> pd.DataFrame:
        """ Method concatenating and returning a logger dataframe with values of the design variables and objective
        function results.

        :return: a logger dataframe
        """
        if self.logs:
            return pd.concat(self.logs).reset_index(drop=True)

    def save_logger(self) -> None:
        """ Method saving logger as a csv file

        """
        if self.output_subdirectory_dir:
            output_table_path = os.path.join(self.output_subdirectory_dir, self.logger_rel_path)
            self.get_logger_df().to_csv(output_table_path)

    def get_mean_fitness_per_generation(self, n_pop) -> pd.DataFrame:
        """ Method calculating the mean fitness per each generation and returning a dataframe with one row per each
        generation. Penalized objective values are excluded from the calculation.

        :param logger_df: full logger dataframe with each row containing information about an individual
        :return: dataframe with one row per each generation of average fitness
        """
        if not self.logs:
            return pd.DataFrame()

        logger_df = self.get_logger_df()
        return self.calculate_mean_of_rows_from_logger_df(logger_df, n_pop)

    @staticmethod
    def calculate_mean_of_rows_from_logger_df(logger_df, n_pop):
        mean_logger_dfs = []
        for index in range(0, len(logger_df), n_pop):
            sub_logger_df = logger_df[(logger_df.index >= index) & (logger_df.index < index + n_pop)]
            sub_logger_df = sub_logger_df[sub_logger_df[SCORE_KEYWORD] < PENALTY]
            mean_logger_dfs.append(sub_logger_df[SCORE_KEYWORD].mean())
        return pd.DataFrame(mean_logger_dfs, columns=[SCORE_KEYWORD])

    def get_min_fitness_per_generation(self, n_pop) -> pd.DataFrame:
        """ Method calculating the min fitness per each generation and returning a dataframe with one row per each
        generation.

        :param logger_df: full logger dataframe with each row containing information about an individual
        :return: dataframe with one row per each generation of minimum fitness
        """
        if not self.logs:
            return pd.DataFrame()

        logger_df = self.get_logger_df()
        return Logger.extract_min_rows_from_logger_df(logger_df, n_pop)

    @staticmethod
    def extract_min_rows_from_logger_df(logger_df, n_pop):
        min_logger_dfs = []
        for index in range(0, len(logger_df), n_pop):
            sub_logger_df = logger_df[(logger_df.index >= index) & (logger_df.index < index + n_pop)]
            idx_min = sub_logger_df[SCORE_KEYWORD].idxmin()
            min_logger_dfs.append(logger_df[logger_df.index == idx_min])
        return pd.concat(min_logger_dfs).reset_index()



import os
from pathlib import Path
import pandas as pd


def read_csv_as_pd(relative_full_file_path):
    full_path = create_resources_file_path(relative_full_file_path)
    return pd.read_csv(full_path, index_col=0)


def create_resources_file_path(relative_full_file_path):
    path = Path(os.path.dirname(__file__))
    full_path = os.path.join(path, relative_full_file_path)
    return full_path



import pandas as pd

from magnumapi.tool_adapters.roxie import RoxieAPI


class RoxieInputBuilder:
    """ Class RoxieInputBuilder builds a ROXIE input

    """

    comment = ''
    bhdata_path = ''
    cadata_path = ''
    iron_path = ''
    flags = {'LEND': False, 'LWEDG': False, 'LPERS': False, 'LQUENCH': False, 'LALGO': False, 'LMIRIRON': False,
             'LBEMFEM': False, 'LPSI': False, 'LSOLV': False, 'LIRON': False, 'LMORPH': False, 'LHARD': False,
             'LPOSTP': False, 'LPEAK': True, 'LINMARG': False, 'LMARG': True, 'LSELF': False, 'LMQE': False,
             'LINDU': False, 'LEDDY': False, 'LSOLE': False, 'LFIELD3': False, 'LFISTR': False, 'LSELF3': False,
             'LBRICK': False, 'LLEAD': False, 'LVRML': False, 'LOPERA': False, 'LOPER20': False, 'LANSYS': False,
             'LRX2ANS': False, 'LANS2RX': False, 'LDXF': False, 'LMAP2D': False, 'LMAP3D': False, 'LEXPR': False,
             'LFIL3D': False, 'LFIL2D': False, 'LCNC': False, 'LANSYSCN': False, 'LWEIRON': False, 'LCATIA': False,
             'LEXEL': False, 'LFORCE2D': False, 'LAXIS': True, 'LIMAGX': False, 'LIMAGY': False, 'LRAEND': False,
             'LMARKER': False, 'LROLER2': False, 'LROLERP': False, 'LIMAGZZ': False, 'LSTEP': False, 'LIFF': False,
             'LICCA': False, 'LICC': False, 'LICCIND': False, 'LITERNL': False, 'LTOPO': False, 'LQUEN3': False,
             'LAYER': True, 'LEULER': True, 'LHEAD': True, 'LPLOT': True, 'LVERS52': True, 'LHARM': True,
             'LMATRF': False, 'LF3LIN': False}
    global2doption = pd.DataFrame()
    global3d = pd.DataFrame()
    block = pd.DataFrame()
    blockoption = pd.DataFrame()
    block3d = pd.DataFrame()
    lead = pd.DataFrame()
    brick = pd.DataFrame()
    iron_yoke = pd.DataFrame()
    extrusion = pd.DataFrame()
    permanentmag2 = pd.DataFrame()
    permanentmag1 = pd.DataFrame()
    layer = pd.DataFrame()
    algo = pd.DataFrame(data=[12], columns=['algo'], index=[0])
    design = pd.DataFrame()
    euler = pd.DataFrame()
    peak = pd.DataFrame()
    timetable2 = pd.DataFrame()
    timetable1 = pd.DataFrame()
    eddy = pd.DataFrame(data=[0], columns=['pers'], index=[0])
    eddyoptions = pd.DataFrame()
    quenchg = pd.DataFrame()
    quenchen = pd.DataFrame()
    quenchtm = pd.DataFrame()
    quenchp = pd.DataFrame()
    quenchs = pd.DataFrame()
    harmonictable = pd.DataFrame([{'no': 1, 'type': 1, 'x': 0, 'y': 0, 's1': 0, 's2': 0,
                                  'rref': 16.667, 'bref': 2, 'field': 0}])
    matrf = pd.DataFrame([{'start': 0, 'end': 0, 'no': 0, 'field': ''},
                          {'start': 0, 'end': 0, 'no': 0, 'field': ''},
                          {'start': 0, 'end': 0, 'no': 0, 'field': 0}])
    linefield = pd.DataFrame()
    harmonicoption = pd.DataFrame()
    graph = pd.DataFrame()
    graphoption = pd.DataFrame()
    plot2d = pd.DataFrame()
    plot2doption = pd.DataFrame()
    plot3d = pd.DataFrame()
    plot3doption = pd.DataFrame()
    objective = pd.DataFrame()

    def set_flag(self, flag_name: str, flag_value: bool) -> "RoxieInputBuilder":
        """ Method setting a flag in a ROXIE input file. An error is thrown if a flag does not exist

        :param flag_name: name of a flag
        :param flag_value: value of a flag
        :return: an updated RoxieInputBuilder instance
        """
        if flag_name in self.flags.keys():
            self.flags[flag_name] = flag_value
        else:
            raise KeyError('Key')
        return self

    def build(self, output_path: str) -> None:
        """ Method building a ROXIE input based on a template file

        :param output_path: an output path for the input .data file
        """
        output_str = self.prepare_data_file_str()

        with open(output_path, 'wb') as input_file:
            input_file.write(bytes(output_str, 'utf-8').replace(b'\r\n', b'\n'))

    def prepare_data_file_str(self) -> str:
        """ Method preparing a data file string content

        :return: a string with lines for an input data file
        """
        outputs = ['VERSION 10.2.1',
                   '\'%s\'' % self.comment,
                   '\'%s\'' % self.bhdata_path,
                   '\'%s\'' % self.cadata_path,
                   '\'%s\'' % self.iron_path,
                   '\n'
                   '&OPTION']
        # Add OPTION
        outputs.append(RoxieInputBuilder.convert_flag_dct_to_str(self.flags))
        # Add GLOBAL2DOPTION
        RoxieInputBuilder.append_to_outputs(outputs, 'GLOBAL2DOPTION', self.global2doption)
        # Add GLOBAL3D
        RoxieInputBuilder.append_to_outputs(outputs, 'GLOBAL3D', self.global3d)
        # Add BLOCK
        RoxieInputBuilder.append_to_outputs(outputs, 'BLOCK', self.block)
        # Add BLOCKOPTION
        RoxieInputBuilder.append_to_outputs(outputs, 'BLOCKOPTION', self.blockoption)
        # Add BLOCK3D
        RoxieInputBuilder.append_to_outputs(outputs, 'BLOCK3D', self.block3d)
        # Add LEAD
        RoxieInputBuilder.append_to_outputs(outputs, 'LEAD', self.lead)
        # Add BRICK
        RoxieInputBuilder.append_to_outputs(outputs, 'BRICK', self.brick)
        # Add IRONYOKE
        RoxieInputBuilder.append_to_outputs(outputs, 'IRONYOKE', self.iron_yoke)
        # Add EXTRUSION
        RoxieInputBuilder.append_to_outputs(outputs, 'EXTRUSION', self.extrusion)
        # Add PERMANENTMAG2
        RoxieInputBuilder.append_to_outputs(outputs, 'PERMANENTMAG2', self.permanentmag2)
        # Add PERMANENTMAG1
        RoxieInputBuilder.append_to_outputs(outputs, 'PERMANENTMAG1', self.permanentmag1)
        # Add LAYER
        RoxieInputBuilder.append_to_outputs(outputs, 'LAYER', self.layer, line_suffix=' /')
        # Add ALGO
        RoxieInputBuilder.append_to_outputs(outputs, 'ALGO', self.algo, header_suffix=' /')
        # Add DESIGN
        RoxieInputBuilder.append_to_outputs(outputs, 'DESIGN', self.design)
        # Add EULER
        RoxieInputBuilder.append_to_outputs(outputs, 'EULER', self.euler)
        # Add PEAK
        self.peak = pd.DataFrame(data=[self.block['no'].values], columns=['']*len(self.block), index=[0])
        RoxieInputBuilder.append_to_outputs(outputs, 'PEAK', self.peak, header_suffix=str(len(self.block['no'])))
        # Add TIMETABLE2
        RoxieInputBuilder.append_to_outputs(outputs, 'TIMETABLE2', self.timetable2)
        # Add TIMETABLE1
        RoxieInputBuilder.append_to_outputs(outputs, 'TIMETABLE1', self.timetable1)
        # Add EDDY
        RoxieInputBuilder.append_to_outputs(outputs, 'EDDY', self.eddy, header_suffix=' /')
        # Add EDDYOPTIONS
        RoxieInputBuilder.append_to_outputs(outputs, 'EDDYOPTIONS', self.eddyoptions)
        # Add QUENCHG
        RoxieInputBuilder.append_to_outputs(outputs, 'QUENCHG', self.quenchg)
        # Add QUENCHEN
        RoxieInputBuilder.append_to_outputs(outputs, 'QUENCHEN', self.quenchen)
        # Add QUENCHTM
        RoxieInputBuilder.append_to_outputs(outputs, 'QUENCHTM', self.quenchtm)
        # Add QUENCHP
        RoxieInputBuilder.append_to_outputs(outputs, 'QUENCHP', self.quenchp)
        # Add QUENCHS
        RoxieInputBuilder.append_to_outputs(outputs, 'QUENCHS', self.quenchs)
        # Add HARMONICTABLE
        RoxieInputBuilder.append_to_outputs(outputs, 'HARMONICTABLE', self.harmonictable)
        # Add MATRF
        RoxieInputBuilder.append_to_outputs(outputs, 'MATRF', self.matrf, header_suffix='1')
        # Add LINEFIELD
        RoxieInputBuilder.append_to_outputs(outputs, 'LINEFIELD', self.linefield)
        # Add HARMONICOPTION
        RoxieInputBuilder.append_to_outputs(outputs, 'HARMONICOPTION', self.harmonicoption)
        # Add GRAPH
        RoxieInputBuilder.append_to_outputs(outputs, 'GRAPH', self.graph)
        # Add GRAPHOPTION
        RoxieInputBuilder.append_to_outputs(outputs, 'GRAPHOPTION', self.graphoption)
        # PLOT2D
        RoxieInputBuilder.append_to_outputs(outputs, 'PLOT2D', self.plot2d, line_suffix=' /')
        # PLOT2DOPTION
        RoxieInputBuilder.append_to_outputs(outputs, 'PLOT2DOPTION', self.plot2doption)
        # PLOT3D
        RoxieInputBuilder.append_to_outputs(outputs, 'PLOT3D', self.plot3d)
        # PLOT3DOPTION
        RoxieInputBuilder.append_to_outputs(outputs, 'PLOT3DOPTION', self.plot3doption)
        # OBJECTIVE
        RoxieInputBuilder.append_to_outputs(outputs, 'OBJECTIVE', self.objective)

        outputs.append('')
        return '\n'.join(outputs)

    @staticmethod
    def convert_flag_dct_to_str(flags: dict) -> str:
        """ Static method converting a dictionary with flags into a formatted string

        :param flags: a dictionary with flags
        :return: a formatted string representation of the dictionary with flags
        """
        COLUMN_WIDTH = 11
        flag_per_line_count = 1
        flag_str = '  '
        for key, value in flags.items():
            temp = "%s=%s" % (key, "T" if value else "F")
            temp += (COLUMN_WIDTH - len(temp)) * ' '
            if flag_per_line_count < 6:
                flag_str += temp
                flag_per_line_count += 1
            else:
                flag_str += temp + "\n  "
                flag_per_line_count = 1

        flag_str += '\n  /'
        return flag_str

    @staticmethod
    def append_to_outputs(outputs, keyword, df, line_suffix='', header_suffix='') -> None:
        """ Static method appending to the output list (in place) a dataframe with a keyword

        :param outputs: a list of outputs for an input file
        :param keyword: a table keyword
        :param df: a dataframe
        """
        outputs.append('')
        outputs.append(RoxieAPI.convert_bottom_header_table_to_str(df, keyword, line_suffix, header_suffix))



import unittest

import numpy as np

from magnumapi.matpro.resistivities import calc_rho_cu_nist
from tests.resource_files import create_resources_file_path


class Test_heat_capacities(unittest.TestCase):
    def test_calc_rho_cu_nist(self):
        # arrange
        T_0 = 1.9
        T = np.linspace(T_0, 300, 100)

        # act
        rho_cu_nist = calc_rho_cu_nist(T, B=5, RRR=100)

        # assert
        path = create_resources_file_path('resources/matpro/resistivities/rho_cu_nist.txt')
        rho_cu_nist_ref = np.loadtxt(path)
        np.testing.assert_allclose(rho_cu_nist_ref, rho_cu_nist)


if __name__ == '__main__':
    unittest.main()




from dataclasses import dataclass

from magnumapi.geometry.definitions.BlockDefinition import BlockDefinition


@dataclass
class HomogenizedBlockDefinition(BlockDefinition):
    """Class for homogenized cos-theta block definition.
       The angle numbering is as follows:
            # 3 ----- 2
            # |       |
            # |       |
            # 0-------1

    Attributes:
       radius_inner (float): The inner radius in millimeter.
       radius_outer (float): The outer radius in millimeter.
       phi_0 (float): The first angle in degrees.
       phi_1 (float): The second angle in degrees.
       phi_2 (float): The third angle in degrees.
       phi_3 (float): The fourth angle in degrees.
    """
    radius_inner: float
    radius_outer: float
    phi_0: float
    phi_1: float
    phi_2: float
    phi_3: float



import math
from dataclasses import dataclass

from magnumapi.cadata.Definition import Definition


@dataclass
class StrandDefinition(Definition):
    """Class for strand definition.

       Attributes:
           d_strand (float): The strand diameter (mm).
           f_cu_nocu (float): The Copper to superconductor ratio.
           rrr (float): The Residual Resistivity ratio.
           temp_ref (float): The Reference Temperature (K) for Bref. This value is never used in the code and serves
           only for reference for Bref and Jc@BrTr.
           b_ref (float): The Reference Field for the definition of a linear approximation of the Jc curve.
           j_c_at_b_ref_t_ref (float): The Critical current density at Bref for the definition of a linear approximation
           of the Jc curve. (LINMARG)
           dj_c_over_db (float): dJc/dB at Bref and Tref for the definition of a linear approximation of the Jc curve.
           (LINMARG)
    """
    d_strand: float
    f_cu_nocu: float
    rrr: float
    temp_ref: float
    b_ref: float
    j_c_at_b_ref_t_ref: float
    dj_c_over_db: float

    @staticmethod
    def get_magnum_to_roxie_dct() -> dict:
        return {"name": "Name",
                "d_strand": "diam.",
                "f_cu_nocu": "cu/sc",
                "rrr": "RRR",
                "temp_ref": "Tref",
                "b_ref": "Bref",
                "j_c_at_b_ref_t_ref": "Jc@BrTr",
                "dj_c_over_db": "dJc/dB",
                "comment": "Comment"}

    def compute_surface_cu(self):
        """ Method computing copper surface of a strand as S = f_co * pi * d_strand^2 / 4

        :return: a copper surface of a strand in mm^2
        """
        return self.compute_f_cu() * self.compute_surface()

    def compute_surface_nocu(self):
        """ Method computing non-copper surface of a strand as S = f_nocu * pi * d_strand^2 / 4

        :return: a non-copper surface of a strand in mm^2
        """
        return self.compute_f_nocu() * self.compute_surface()

    def compute_f_nocu(self):
        """ Method fraction of non-copper in a strand

        :return: a fraction of non-copper in a strand (no unit)
        """
        return 1 / (1 + self.f_cu_nocu)

    def compute_f_cu(self):
        """ Method fraction of copper in a strand

        :return: a fraction of copper in a strand (no unit)
        """
        return self.f_cu_nocu / (1 + self.f_cu_nocu)

    def compute_surface(self):
        """ Method computing surface of a strand as S = pi * d_strand^2 / 4

        :return: a surface of a strand in mm^2
        """
        return math.pi * self.d_strand**2 / 4



import os
from distutils.dir_util import copy_tree
from typing import List

from magnumapi.optimization.OptimizationNotebookConfig import OptimizationNotebookConfig


class DirectoryManager:
    """ A DirectoryManager class providing basic functionalities of checking file presence, creating directories, etc.

    """

    @staticmethod
    def check_if_file_exists(file_path: str) -> None:
        """Static method checking whether a file exists. If not, then a FileNotFoundError is raised.

        :param file_path: a path to a file whose presence is verified.
        """
        if not os.path.isfile(file_path):
            raise FileNotFoundError('The file %s does not exist!' % file_path)

    @staticmethod
    def create_directory_if_nonexistent(output_dir: str) -> None:
        """ Static method checking whether a directory exists, if not then it is created

        :param output_dir: a path to an output directory
        """
        is_dir = os.path.isdir(output_dir)
        if not is_dir:
            os.mkdir(output_dir)

    @staticmethod
    def find_output_subdirectory_name(output_dir: str) -> str:
        """ Method finding an output subdirectory index by searching the maximum index present in an output directory
        and incrementing that index.

        :param output_dir: a path to a root output directory
        :return: a name of an output subdirectory corresponding to the incremented index
        """
        current_output_folder = '1'
        int_folder_names = [int(name) for name in os.listdir(output_dir) if name.isnumeric()]
        if int_folder_names:
            max_folder_name = max(int_folder_names)
            current_output_folder = str(max_folder_name + 1)

        return current_output_folder

    @classmethod
    def create_output_subdirectory(cls, output_dir: str, subdirectory_name: str) -> None:
        """ Static method creating an output subdirectory

        :param output_dir:
        :param subdirectory_name:
        """
        output_subdirectory_dir = os.path.join(output_dir, subdirectory_name)
        cls.create_directory_if_nonexistent(output_subdirectory_dir)

    @classmethod
    def copy_notebook_folders(cls,
                              input_folder: str,
                              notebook_configs: List[OptimizationNotebookConfig],
                              output_subdirectory_dir) -> None:
        """ Static method copying folders containing notebooks from the source directory to the output subdirectory

        :param input_folder: a root input folder
        :param notebook_configs: a list of notebook configs
        :param output_subdirectory_dir: a path to an output subdirectory
        """
        for notebook_config in notebook_configs:
            notebook_folder = notebook_config.notebook_folder
            input_notebook_folder_dir = os.path.join(input_folder, notebook_folder)
            output_notebook_subdirectory_dir = os.path.join(output_subdirectory_dir, notebook_folder)
            cls.create_directory_if_nonexistent(output_notebook_subdirectory_dir)
            copy_tree(input_notebook_folder_dir, output_notebook_subdirectory_dir)

    @classmethod
    def copy_model_input(cls, input_folder_path: str, input_folder_name: str, output_subdirectory_dir: str) -> None:
        """ Static method copying a model input from an input directory to an output subdirectory

        :param input_folder_path:
        :param input_folder_name:
        :param output_subdirectory_dir:
        """
        # ToDo: should be renamed to better denote the purpose
        ref_input_folder = os.path.join(input_folder_path, input_folder_name)
        new_input_folder = os.path.join(output_subdirectory_dir, input_folder_name)

        cls.create_directory_if_nonexistent(new_input_folder)
        copy_tree(ref_input_folder, new_input_folder)



from itertools import accumulate
from typing import List

import pandas as pd

import plotly.graph_objects as go
from plotly.subplots import make_subplots

from magnumapi.geometry.GeometryPlot import GeometryPlot
from magnumapi.optimization.DesignVariable import CockpitGeneticDesignVariable, GeneticDesignVariable
from magnumapi.optimization.GeneticOptimization import RoxieGeneticOptimization
from magnumapi.optimization.Logger import Logger
from magnumapi.optimization.OptimizationConfig import OptimizationConfig
from magnumapi.optimization.constants import SCORE_KEYWORD


class OptimizationCockpitWidget:
    """ A class providing a plotly widget for an optimization cockpit. The cockpit displays:
    - progress of the optimization
    - objective variable table
    - comparison of fitness function of a selected individual to the current best one
    - comparison of design variables of a selected individual to the current best one
    - selected geometry overview
    The cockpit is interactive and allows for selecting an individual to display its geometry

    """

    def __init__(self,
                 gen_opt: RoxieGeneticOptimization,
                 logger_df: pd.DataFrame,
                 config: OptimizationConfig,
                 design_variables_df: pd.DataFrame,
                 is_log_y=False) -> None:
        """ Constructor of an OptimizationCockpitWidget instance

        :param gen_opt: genetic optimization instance
        :param logger_df: a dataframe with logger
        :param config: optimization config instance
        :param design_variables_df: a dataframe with objective variables
        :param is_log_y: if True, than the score plot has logarithmic y-scale
        """
        self.gen_opt = gen_opt
        self.min_logger_df = Logger.extract_min_rows_from_logger_df(logger_df, self.gen_opt.n_pop)
        self.config = config
        self.design_variables = initialize_design_variables(design_variables_df)
        self.is_log_y = is_log_y

        fig = make_subplots(
            rows=3, cols=2,
            shared_xaxes=False,
            vertical_spacing=0.03,
            column_widths=[0.35, 0.65],
            row_heights=[0.15, 0.2, 0.65],
            specs=[[{"type": "xy"}, {"type": "table"}], [{"type": "xy", "rowspan": 2}, {"type": "scatter"}],
                   [None, {"type": "scatter"}]]
        )

        self.widget = go.FigureWidget(fig)

    def show(self) -> go.FigureWidget:
        """ Method displaying the widget by returning the field that contains it.

        :return: optimization cockpit widget
        """
        return self.widget

    def build(self) -> None:
        """ Method building the optimization cockpit widget with plotly library

        """
        self.update_title(len(self.min_logger_df), self.min_logger_df.loc[self.min_logger_df.index[-1], SCORE_KEYWORD])

        # Display fitness function graph with callback
        self.display_fitness_function_graph(row=2, col=2)

        index_best = self.min_logger_df.index[-1]

        # Display selected generation objective table
        obj_best_df = create_objective_table(self.config, self.min_logger_df, index_best)
        self.display_objective_table(obj_best_df)

        # Display fitness function comparison graph
        self.display_fitness_function_comparison_graph(obj_best_df, obj_best_df, row=1, col=1)

        # Display design variables graph
        self.display_design_variables_graph(index_best, index_best, row=2, col=1)

        # Display geometry graph
        self.display_geometry_graph(index_best, row=3, col=2)

        # Create callback function
        def callback(object: "OptimizationCockpitWidget"):
            def update_point(trace, points, selector):
                index = points.point_inds[0]

                # Update selected generation objective table
                widget_data = object.widget.data[1]
                obj_df = create_objective_table(object.config, object.min_logger_df, index)
                OptimizationCockpitWidget.update_objective_table(widget_data, obj_df)

                # Clear remaining graphs before display
                object.widget.data = [object.widget.data[0], object.widget.data[1]]

                # Display fitness function comparison graph
                object.display_fitness_function_comparison_graph(obj_df, obj_best_df, row=1, col=1)

                # Display design variables graph
                object.display_design_variables_graph(index, index_best, row=2, col=1)

                # Display geometry graph
                object.display_geometry_graph(index, row=3, col=2)

                object.update_title(index, object.min_logger_df.loc[index, SCORE_KEYWORD])

            return update_point

        scatter = self.widget.data[0]
        self.widget.layout.hovermode = 'closest'
        scatter.on_click(callback(self))

    def update_title(self, index_gen: int, fitness_function: float) -> None:
        """ Method updating the widget title with generation index and fitness function value

        :param index_gen: generation index (0-based)
        :param fitness_function: fitness function value
        """
        self.widget.update_layout(
            height=1000,
            showlegend=False,
            title_x=0.5,
            title_text="Optimization Cockpit - index: %d - fitness function: %f" % (index_gen, fitness_function),
        )

    def display_fitness_function_graph(self, row=2, col=2) -> None:
        """ Method displaying the fitness function graph in the widget at a given index (row and column)

        :param row: row index of the graph (1-based)
        :param col: column index of the graph (1-based)
        """
        self.widget.add_trace(
            go.Scatter(
                x=self.min_logger_df.index.values,
                y=self.min_logger_df[SCORE_KEYWORD].values,
                mode="lines+markers",
                name="fitness",
            ),
            row=row, col=col
        )
        if self.is_log_y:
            self.widget.update_yaxes(type='log', row=row, col=col)

    def display_objective_table(self, obj_df: pd.DataFrame) -> None:
        """ Method displaying the objective table in the widget

        :param obj_df: a dataframe with information to be displayed
        """
        obj_trans_df = transpose_objective_table(obj_df)

        self.widget.add_trace(
            go.Table(
                header=dict(
                    values=obj_trans_df.columns,
                    font=dict(size=10),
                    align="left"
                ),
                cells=dict(
                    values=[obj_trans_df[k].tolist() for k in obj_trans_df.columns],
                    align="left")
            ),
            row=1, col=2)

    @staticmethod
    def update_objective_table(widget_data, obj_df: pd.DataFrame) -> None:
        """ Static method updating the objective table

        :param widget_data: reference to the table widget in the cockpit
        :param obj_df: new objective dataframe
        """
        obj_trans_df = transpose_objective_table(obj_df)

        widget_data.cells = dict(
            values=[obj_trans_df[k].tolist() for k in obj_trans_df.columns],
            align="left")

    def display_fitness_function_comparison_graph(self,
                                                  obj_df: pd.DataFrame,
                                                  obj_best_df: pd.DataFrame,
                                                  row=1,
                                                  col=1) -> None:
        """ Method displaying a fitness function comparison graph. The bar graph contains contributions of each of the
        objectives.

        :param obj_df: objective table for a selected individual
        :param obj_best_df: objective table for the best individual
        :param row: row index of the graph in the widget (1-based)
        :param col: column index of the graph in the widget (1-based)
        """
        objective_variables = obj_df.index.values
        fitness_actual = obj_df['objective_weighted'].values
        fitness_best = obj_best_df['objective_weighted'].values

        cum_fitness_actual = list(accumulate(fitness_actual))
        cum_fitness_best = list(accumulate(fitness_best))

        hover_text_actual = ['%s: %f' % obj_fitness for obj_fitness in zip(objective_variables, fitness_actual)]
        hover_text_best = ['%s: %f' % obj_fitness for obj_fitness in zip(objective_variables, fitness_best)]

        n = len(objective_variables)
        for i in range(n):
            trace = go.Bar(
                x=[cum_fitness_best[n - i - 1], cum_fitness_actual[n - i - 1]],
                y=['Best', 'Actual'],
                hovertext=[hover_text_best[n - i - 1], hover_text_actual[n - i - 1]],
                orientation='h',
                hoverinfo='text',
                offsetgroup=1
            )
            self.widget.append_trace(trace, row, col)

    def display_design_variables_graph(self, index_act: int, index_best: int, row=2, col=1) -> None:
        """ Method displaying the design variables graph. The graph compares a selected individual with the best one.

        :param index_act: index of a selected individual (0-based)
        :param index_best: index of the best individual (0-based)
        :param row: row index of the graph in the widget (1-based)
        :param col: column index of the graph in the widget (1-based)
        """

        trace_best = self._prepare_bar_plot_with_fraction(index_best, 'Best')
        trace_act = self._prepare_bar_plot_with_fraction(index_act, 'Actual')

        self.widget.add_traces([trace_best, trace_act], rows=row, cols=col)

    def _prepare_bar_plot_with_fraction(self, index: int, name: str) -> go.Bar:
        return go.Bar(x=prepare_fractions_for_bar_plot(self.design_variables, self.min_logger_df, index),
                      hovertext=prepare_hover_texts_for_bar_plot(self.design_variables, self.min_logger_df, index),
                      y=[dv.variable_name for dv in self.design_variables],
                      name=name,
                      orientation='h')

    def display_geometry_graph(self, index: int, row=3, col=2) -> None:
        """ Method displaying a geometry preview for a selected index.

        :param index: index of a selected individual
        :param row: row index of the graph in the widget (1-based)
        :param col: column index of the graph in the widget (1-based)
        """
        fit_row = self.min_logger_df[self.min_logger_df.index == index]
        fit_dct = fit_row.to_dict('records')[0]
        if self.gen_opt.config.model_creation_type == 'default':
            geometry = self.gen_opt.update_model_parameters(fit_dct)
        elif self.gen_opt.config.model_creation_type == 'targeted':
            geometry = self.gen_opt.update_model_parameters_targeted(fit_dct)
        elif self.gen_opt.config.model_creation_type == 'programmable':
            geometry = self.gen_opt.update_model_parameters_targeted(fit_dct)
        else:
            raise AttributeError('Model creation type %s not supported!' % self.config.model_creation_type)
        geometry.build_blocks()

        go_scatter = GeometryPlot._create_plotly_scatters(geometry.blocks)

        # To avoid plotting issues in case the number of turns
        max_number_of_blocks = int(sum([dv.xu for dv in self.design_variables if dv.variable_name == 'nco']))
        for i in range(max_number_of_blocks - len(go_scatter)):
            go_scatter.append(go.Scatter(x=[], y=[]))

        self.widget.add_traces(go_scatter, rows=row, cols=col)


# Testable support functions
def initialize_design_variables(design_variables_df: pd.DataFrame) -> List[CockpitGeneticDesignVariable]:
    """ Static method initializing genetic optimization cockpit design variables from a DataFrame

    :param design_variables_df: a DataFrame with design variables
    :return: a list of design variables
    """
    design_variables = []
    for index, row in design_variables_df.iterrows():
        if 'layer' not in row['variable']:
            design_variables.append(CockpitGeneticDesignVariable(GeneticDesignVariable(**row.to_dict())))

    return design_variables


def create_objective_table(config: OptimizationConfig,
                           min_logger_df: pd.DataFrame,
                           index: int
                           ) -> pd.DataFrame:
    """ Method creating the objective table by extracting weights, constraints, and weighted objectives

    :param config:
    :param min_logger_df:
    :param index: index of a selected best individual per generation
    :return: a dataframe representing the selected individual
    """
    objectives = [obj.objective for obj in config.objectives]
    obj_df = pd.DataFrame(min_logger_df.loc[index, objectives])
    obj_df = obj_df.rename(columns={obj_df.columns[0]: 'objective'})

    obj_df['weights'] = obj_df.apply(lambda col: config.get_weight(col.name), axis=1)
    obj_df['constraints'] = obj_df.apply(lambda col: config.get_constraint(col.name), axis=1)
    obj_df['objective_weighted'] = obj_df['weights'] * (obj_df['objective'] - obj_df['constraints'])

    return obj_df


def transpose_objective_table(obj_df: pd.DataFrame) -> pd.DataFrame:
    """ Static method transposing the objective table for a selected best individual for a generation

    :param obj_df: input objective table
    :return: transposed input objective table
    """
    obj_trans_df = obj_df.T
    obj_trans_df = obj_trans_df.reset_index()
    obj_trans_df = obj_trans_df.rename(columns={"index": ""})
    return obj_trans_df


def prepare_fractions_for_bar_plot(design_variables: List[CockpitGeneticDesignVariable],
                                   min_logger_df: pd.DataFrame,
                                   index: int
                                   ) -> List[float]:
    """ Prepare a list of fractions of design variable values for a bar plot display

    :param design_variables: a list of cockpit design variables
    :param min_logger_df:
    :param index:
    :return: a list of fractions for a bar plot
    """
    return [dv.get_fraction(min_logger_df.loc[index, dv.logger_column]) for dv in design_variables]


def prepare_hover_texts_for_bar_plot(design_variables: List[CockpitGeneticDesignVariable],
                                     min_logger_df: pd.DataFrame,
                                     index: int
                                     ) -> List[float]:
    """ Prepare a list of hover texts for a bar plot display

    :param design_variables: a list of cockpit design variables
    :param min_logger_df:
    :param index:
    :return:a list of hover texts for a bar plot
    """
    return [dv.get_hover_text(min_logger_df.loc[index, dv.logger_column]) for dv in design_variables]



from dataclasses import dataclass

from magnumapi.cadata.Definition import Definition


@dataclass
class ConductorDefinition(Definition):
    """Class for conductor definition.

       Attributes:
           type (int): The conductor type (e.g., HTS tape, Rutherford cable).
           cable_geom (str): The cable geometry name (defined in Cable geometry block).
           strand (str): The strand name (defined in Strand block).
           filament (str): The filament name (defined in Filament block).
           insulation (str): The insulation name (defined in Insul block).
           transient (str): -
           quench_mat (str): Quench material, not always present. Defaults to an empty string.
           temp_ref (float): The operating temperature in K. All conductors in the cross-section must have the same
            temperature. Adaptations can be made via the design variables.
    """
    type: int
    cable_geom: str
    strand: str
    filament: str
    insulation: str
    transient: str
    temp_ref: float
    quench_mat: str = ''

    @staticmethod
    def get_magnum_to_roxie_dct() -> dict:
        return {"name": "Name",
                "type": "Type",
                "cable_geom": "CableGeom.",
                "strand": "Strand",
                "filament": "Filament",
                "insulation": "Insul",
                "transient": "Trans",
                "quench_mat": "QuenchMat.",
                "temp_ref": "T_o",
                "comment": "Comment"}





def read(file_path: str) -> str:
    #todo: add check if the file exists - as a decorator
    with open(file_path, 'r') as file:
        return file.read()


def readlines(file_path: str) -> list:
    #todo: add check if the file exists - as a decorator
    with open(file_path, 'r') as file:
        return file.readlines()


def write(file_path: str, content: str) -> None:
    #todo: add check if the file exists - as a decorator
    with open(file_path, 'w') as file:
        file.write(content)


def writelines(file_path: str, contents: list, endline='\n') -> None:
    #todo: add check if the file exists - as a decorator
    with open(file_path, 'w') as file:
        for content in contents:
            file.write(content + endline)


import pandas as pd
import matplotlib.pyplot as plt

from magnumapi.geometry.primitives.Area import Area
from magnumapi.geometry.primitives.Point import Point
from magnumapi.geometry.blocks.Block import Block
from magnumapi.cadata.CableDefinition import CableDefinition
from magnumapi.cadata.ConductorDefinition import ConductorDefinition
from magnumapi.cadata.InsulationDefinition import InsulationDefinition
from magnumapi.geometry.definitions.RectangularBlockDefinition import RectangularBlockDefinition
from magnumapi.geometry.definitions.HomogenizedRectangularBlockDefinition import HomogenizedRectangularBlockDefinition
from magnumapi.cadata.StrandDefinition import StrandDefinition


class RectangularBlock(Block):
    """ RectangularBlock class providing a container and construction method for blocks composed of rectangular turns

    """

    # A dictionary mapping from a ROXIE to magnum naming convention for block definition
    roxie_to_magnum_dct = {"radius": "x",
                           "phi": "y"}

    def __init__(self,
                 block_def: RectangularBlockDefinition,
                 cable_def: CableDefinition,
                 insul_def: InsulationDefinition,
                 strand_def: StrandDefinition,
                 conductor_def: ConductorDefinition) -> None:
        """ Constructor of the RectangularBlock class

        An attribute error is raised in case cable definition contains a trapezoidal cable.

        :param block_def: a rectangular block definition with x,y coordinates
        :param cable_def: a cable definition from cadata
        :param insul_def: an insulation definition from cadata
        :param strand_def: a strand definition from cadata
        :param conductor_def: a conductor definition from cadata
        """
        super().__init__(cable_def=cable_def,
                         insul_def=insul_def,
                         strand_def=strand_def,
                         conductor_def=conductor_def)
        self.block_def = block_def

        if self.cable_def.thickness_i != self.cable_def.thickness_o:
            raise AttributeError('Rectangular blocks do not work with trapezoidal cables '
                                 '(thickness_i = %.4f, thickness_o = %.4f).' %
                                 (self.cable_def.thickness_i, self.cable_def.thickness_o))

    def build_block(self) -> None:
        p_shift = Point.of_cartesian(self.block_def.x, self.block_def.y)

        for i in range(self.block_def.nco):
            p1 = Point.of_cartesian(0.0, 0.0)
            p2 = Point.of_cartesian(self.cable_def.width + 2 * self.insul_def.width, 0.0)
            p3 = Point.of_cartesian(self.cable_def.width + 2 * self.insul_def.width,
                                    self.cable_def.thickness_i + 2 * self.insul_def.thickness)
            p4 = Point.of_cartesian(0.0, self.cable_def.thickness_i + 2 * self.insul_def.thickness)
            area = Area.of_closed_ordered_list_of_points(points=(p1, p2, p3, p4))
            area = area.rotate(self.block_def.alpha).translate(p_shift)
            self.areas.append(area)

            p_shift = area.get_line(3).p1.copy()

    def plot_block(self, ax: plt.Axes) -> None:
        for area in self.areas:
            area.plot(ax)

    def plot_bare_block(self, ax: plt.Axes) -> None:
        for bare_area in self.get_bare_areas():
            bare_area.plot(ax)

    def to_block_df(self):
        dct_no_areas = self.to_abs_dict()
        df = pd.DataFrame(dct_no_areas, index=[0])
        return df[['no', 'type', 'nco', 'radius', 'phi', 'alpha', 'current', 'condname', 'n1', 'n2', 'imag', 'turn']]

    def to_abs_dict(self):
        dct = self.block_def.__dict__
        dct_no_areas = {key: value for key, value in dct.items()}
        for key, value in self.roxie_to_magnum_dct.items():
            dct_no_areas[key] = dct_no_areas.pop(value)
        return dct_no_areas

    def to_rel_dict(self, alpha_ref=0.0, phi_ref=0.0):
        raise NotImplementedError('This method is not implemented for this class')

    def get_bare_area(self, area: Area) -> Area:
        p1 = area.get_line(0).p1 + Point.of_cartesian(self.insul_def.width,
                                                      self.insul_def.thickness).rotate(self.block_def.alpha)
        p2 = area.get_line(1).p1 + Point.of_cartesian(-self.insul_def.width,
                                                      self.insul_def.thickness).rotate(self.block_def.alpha)
        p3 = area.get_line(2).p1 + Point.of_cartesian(-self.insul_def.width,
                                                      -self.insul_def.thickness).rotate(self.block_def.alpha)
        p4 = area.get_line(3).p1 + Point.of_cartesian(self.insul_def.width,
                                                      -self.insul_def.thickness).rotate(self.block_def.alpha)
        return Area.of_closed_ordered_list_of_points(points=(p1, p2, p3, p4))

    def homogenize(self):
        if not self.areas:
            raise AttributeError('In order to extract a homogenized block definition, areas need to be initialized!')

        # write block corner positions
        # 3 ------- x_up_r, y_up_r
        # |         |
        # |         |
        # x,y------ 1

        x = self.areas[0].get_line(0).p1.x
        y = self.areas[0].get_line(0).p1.y

        x_up_r = self.areas[-1].get_line(2).p1.x
        y_up_r = self.areas[-1].get_line(2).p1.y

        param_dct = dict(no=self.block_def.no, type=1, nco=self.block_def.nco, x=x, y=y, x_up_r=x_up_r, y_up_r=y_up_r,
                         current=self.block_def.current, condname=self.block_def.condname, alpha=self.block_def.alpha,
                         n1=self.block_def.n1, n2=self.block_def.n2, imag=self.block_def.imag,
                         turn=self.block_def.turn)

        return HomogenizedRectangularBlock(block_def=HomogenizedRectangularBlockDefinition(**param_dct),
                                           cable_def=self.cable_def,
                                           strand_def=self.strand_def,
                                           conductor_def=self.conductor_def,
                                           insul_def=self.insul_def)


class HomogenizedRectangularBlock(RectangularBlock):
    """ HomogenizedRectangularBlock class providing a container and construction method for blocks composed of
    homogenized rectangular turns

    """

    def __init__(self,
                 block_def: HomogenizedRectangularBlockDefinition,
                 cable_def: CableDefinition,
                 insul_def: InsulationDefinition,
                 strand_def: StrandDefinition,
                 conductor_def: ConductorDefinition) -> None:
        super().__init__(block_def, cable_def, insul_def, strand_def, conductor_def)
        self.block_def = block_def

    def plot_block(self, ax):
        self.areas[0].plot(ax)

    def to_block_df(self):
        raise NotImplementedError('This method is not implemented for this class')

    def to_abs_dict(self):
        raise NotImplementedError('This method is not implemented for this class')

    def to_rel_dict(self, alpha_ref, phi_ref):
        raise NotImplementedError('This method is not implemented for this class')

    def build_block(self):
        # write block corner positions
        # 3 ------- 2
        # |         |
        # |         |
        # 0 ------- 1
        p0 = Point.of_polar(self.block_def.x, self.block_def.y)
        p1 = Point.of_polar(self.block_def.x_up_r, self.block_def.y)
        p2 = Point.of_polar(self.block_def.x_up_r, self.block_def.y_up_r)
        p3 = Point.of_polar(self.block_def.x, self.block_def.y_up_r)
        self.areas = [Area.of_closed_ordered_list_of_points((p0, p1, p2, p3))]

    def homogenize(self):
        return self



import math
from typing import Union
from enum import Enum, auto

import numpy as np


class DesignVariable:
    """ A DesignVariable class for optimization purposes
    """

    def __init__(self,
                 xl: Union[int, float],
                 xu: Union[int, float],
                 variable: str,
                 layer: int,
                 bcs: str) -> None:
        """ A DesignVariable constructor

        :param xl: lower limit of a design variable
        :param xu: upper limit of a design variable
        :param variable: name of a design variable
        :param layer: layer index; this brakes the compatibility with ROXIE
        :param bcs: block index, range of indices (with -) or no block indication for global variables
        """
        self.xl = xl
        self.xu = xu
        self.variable = variable
        self.layer = layer
        try:
            bcs = str(int(bcs))
        except ValueError:
            bcs = str(bcs)
        self.bcs = bcs

    def get_variable_names(self) -> list:
        """ Method returning a design variable name. There are three output types
        - if bcs string is empty, then the variable name is returned - for global design variables
        - if bcs string contains a dash, then the range of indices is considered
        - if bcs string contains a number, then the variable:bcs name is returned
        - otherwise, an AttributeError is raised

        :return: a list of strings with variable names
        """
        design_variable_type = self.get_variable_type()
        if design_variable_type == DesignVariableType.GLOBAL_VARIABLE:
            return [self.variable]
        elif design_variable_type == DesignVariableType.LAYER_VARIABLE:
            return ["%s:%s:" % (self.variable, self.layer)]
        elif design_variable_type == DesignVariableType.MULTI_TURN_VARIABLE:
            return self._convert_range_of_blocks_into_list()
        elif design_variable_type == DesignVariableType.TURN_VARIABLE:
            return ["%s:%s:%s" % (self.variable, int(self.layer), self.bcs)]
        else:
            raise AttributeError('The design variable has incorrect block index value: %s.' % self.bcs)

    def _convert_range_of_blocks_into_list(self) -> list:
        bcs_ranges = self.bcs.split('-')
        if len(bcs_ranges) != 2:
            raise AttributeError('The block index range definition %s is wrong. Only one hyphen is allowed.' % self.bcs)

        if not bcs_ranges[0].isnumeric():
            raise AttributeError('The lower block index range %s is not a number.' % bcs_ranges[0])

        if not bcs_ranges[1].isnumeric():
            raise AttributeError('The upper block index range %s is not a number.' % bcs_ranges[1])

        lower = int(bcs_ranges[0])
        upper = int(bcs_ranges[1])

        if lower > upper:
            raise AttributeError('The lower index %d is greater than the upper one %d.' % (lower, upper))

        return ['%s:%s:%d' % (self.variable, self.layer, block) for block in range(lower, upper + 1)]

    def get_variable_type(self) -> "DesignVariableType":
        """ Method returning a design variable type: GLOBAL_VARIABLE, LAYER_VARIABLE, MULTI_TURN_VARIABLE, TURN_VARIABLE

        :return: a DesignVariabletype instance
        """
        if (self.bcs == '' or self.bcs == 'nan') and \
                (self.layer == '' or self.layer == 'nan' or math.isnan(self.layer)):
            return DesignVariableType.GLOBAL_VARIABLE
        elif self.bcs == 'nan' or self.bcs == '':
            return DesignVariableType.LAYER_VARIABLE
        elif '-' in self.bcs:
            return DesignVariableType.MULTI_TURN_VARIABLE
        elif self.bcs.isnumeric():
            return DesignVariableType.TURN_VARIABLE
        else:
            raise AttributeError('The design variable has incorrect block index value: %s.' % self.bcs)


class DesignVariableType(Enum):
    GLOBAL_VARIABLE = auto()
    LAYER_VARIABLE = auto()
    MULTI_TURN_VARIABLE = auto()
    TURN_VARIABLE = auto()

class GeneticDesignVariable(DesignVariable):
    """ A DesignVariable class for genetic optimization purpose

    """

    def __init__(self,
                 xl: Union[int, float],
                 xu: Union[int, float],
                 variable: str,
                 layer: int,
                 bcs: str,
                 variable_type: str,
                 bits: int) -> None:
        """ A GeneticDesignVariable constructor

        :param xl: lower limit of a design variable
        :param xu: upper limit of a design variable
        :param variable: name of a design variable
        :param bcs: block index, range of indices (with -) or no block indication for global variables
        :param variable_type: variable type (int, float) used to determine the gene to value conversion type
        :param bits: the number of bits to represent a design variable
        """
        if variable_type == 'int':
            super().__init__(int(xl), int(xu), variable, layer, bcs)
        else:
            super().__init__(float(xl), float(xu), variable, layer, bcs)
        self.bits = bits
        self.variable_type = variable_type

    def generate_random_gene(self) -> list:
        """ Method generating a random gene for int and float design variables

        :return:
        """
        gene = list(np.random.randint(0, 2, self.bits))
        if self.variable_type == 'float':
            return gene
        elif self.variable_type == 'int':
            gene_int = GeneticDesignVariable.convert_gene_to_int(gene)
            if gene_int > (self.xu - self.xl):
                gene_temp = GeneticDesignVariable.convert_int_to_gene(self.xu - self.xl, self.bits)
                return (self.bits - len(gene_temp)) * [0] + gene_temp
            else:
                return gene
        else:
            raise AttributeError('The variable type %s should be either int or float.' % self.variable_type)

    def convert_gene_to_value(self, gene: list) -> Union[int, float]:
        """ Method converting a gene to design variable value (either int or float):
        - for int, the value is given as lower limit + gene integer value; in case of an overflow, the upper limit is
        considered
        - for float, the value is given as the lower limit + gene integer value multiplied by variable range and divided
        by the number of binary combinations

        :param gene: a list of bits representing a design variable
        :return: numeric value of a gene
        """
        gene_int = GeneticDesignVariable.convert_gene_to_int(gene)

        # convert to value
        if self.variable_type == 'int':
            return min(int(self.xl + gene_int), int(self.xu))
        else:
            return self.xl + gene_int * (self.xu - self.xl) / 2 ** self.bits

    @staticmethod
    def convert_gene_to_int(gene: list) -> int:
        """ Static method converting a gene as a list into an integer

        :param gene: a list of bits representing a gene
        :return: an integer value of a gene
        """
        # convert chromosome to a string of chars
        gene_chars = ''.join([str(s) for s in gene])

        # convert string to integer
        return int(gene_chars, 2)

    @staticmethod
    def convert_int_to_gene(n: int, nbits: int) -> list:
        """ Static method converting an integer to a gene

        :param n: an integer value of a gene
        :param nbits: number of bits in binary representation
        :return: a list of bits representing a gene
        """
        bitstring = [n >> i & 1 for i in range(n.bit_length() - 1, -1, -1)]
        if len(bitstring) < nbits:
            bitstring = [0] * (nbits - len(bitstring)) + bitstring
        return bitstring

class CockpitGeneticDesignVariable(GeneticDesignVariable):

    def __init__(self, gen_dv: "GeneticDesignVariable") -> None:
        super().__init__(**gen_dv.__dict__)
        self.logger_column = self.get_minimal_logger_column_name()
        self.variable_name = self.get_variable_name_for_widget_bar_plot()

    def get_minimal_logger_column_name(self) -> str:
        design_variable_type = self.get_variable_type()
        if design_variable_type == DesignVariableType.GLOBAL_VARIABLE:
            return self.variable
        elif design_variable_type == DesignVariableType.LAYER_VARIABLE:
            return "%s:%s:" % (self.variable, int(self.layer))
        elif design_variable_type == DesignVariableType.MULTI_TURN_VARIABLE:
            return '%s:%s:%s' % (self.variable, self.layer, self.bcs.split('-')[0])
        elif design_variable_type == DesignVariableType.TURN_VARIABLE:
            return '%s:%s:%s' % (self.variable, self.layer, self.bcs)
        else:
            raise AttributeError('The design variable has incorrect block index value: %s.' % self.bcs)

    def get_variable_name_for_widget_bar_plot(self) -> str:
        design_variable_type = self.get_variable_type()
        if design_variable_type == DesignVariableType.GLOBAL_VARIABLE:
            return self.variable
        elif design_variable_type == DesignVariableType.LAYER_VARIABLE:
            return "%s:%s:" % (self.variable, int(self.layer))
        elif design_variable_type == DesignVariableType.MULTI_TURN_VARIABLE:
            return '%s:%s:%s' % (self.variable, self.layer, self.bcs)
        elif design_variable_type == DesignVariableType.TURN_VARIABLE:
            return '%s:%s:%s' % (self.variable, int(self.layer), self.bcs)
        else:
            raise AttributeError('The design variable has incorrect block index value: %s.' % self.bcs)

    def get_hover_text(self, value):
        if self.variable_type == 'float':
            return 'value: %.3f, range: [%.3f, %.3f]' % (value, self.xl, self.xu)
        else:
            return 'value: %d, range: [%d, %d]' % (value, self.xl, self.xu)

    def get_fraction(self, value):
        if self.xu == self.xl:
            return float('nan')
        else:
            return (value - self.xl) / (self.xu - self.xl)



from dataclasses import dataclass
from abc import ABC, abstractmethod


@dataclass
class Definition(ABC):
    """Base class for a definition.

       Attributes:
           name (str): The name of a definition (serves as a primary key).
           comment (str): The comment.
    """
    name: str
    comment: str

    @staticmethod
    @abstractmethod
    def get_magnum_to_roxie_dct():
        """ Method returning a mapping from MagNum naming convention to ROXIE naming convention.

        :return: a dictionary from MagNum to ROXIE names
        """
        raise NotImplementedError('This method is not implemented for this class')

    @classmethod
    def get_roxie_to_magnum_dct(cls):
        """ Method returning a mapping from ROXIE naming convention to MagNum naming convention. In case there is a
        field not present in ROXIE but present in MagNum, then its default value is None.

        :return: a dictionary from ROXIE to MagNum names
        """
        return {val: key for key, val in cls.get_magnum_to_roxie_dct().items() if val is not None}

    @classmethod
    def reorder_dct(cls, dct: dict) -> dict:
        """ Method reordering an input dictionary to follow ROXIE ordering and magnum naming convention. The reason for
        this is the use of inheritance which does not follow the ROXIE ordering.

        :param dct:
        :return:
        """
        for key in cls.get_magnum_to_roxie_dct().keys():
            dct[key] = dct.pop(key)
        return dct



from unittest import TestCase

import dataclasses

from magnumapi.geometry.primitives.Area import Area
from magnumapi.geometry.primitives.Point import Point
from magnumapi.geometry.primitives.Line import Line


class TestArea(TestCase):

    def test_get_line(self):
        # arrange
        p1 = Point.of_polar(10, 10)
        p2 = Point.of_polar(20, 30)
        p3 = Point.of_polar(15, 45)


        l1 = Line.of_end_points(p1, p2)
        l2 = Line.of_end_points(p2, p3)
        l3 = Line.of_end_points(p3, p1)

        area = Area.of_lines((l1, l2, l3))

        # act
        # assert
        with self.assertRaises(IndexError) as context:
            area.get_line(5)

        self.assertTrue('The requested line index 5 is out of range for a list of length 3' in str(context.exception))

    def test_get_line_negative(self):
        # arrange
        p1 = Point.of_polar(10, 10)
        p2 = Point.of_polar(20, 30)
        p3 = Point.of_polar(15, 45)


        l1 = Line.of_end_points(p1, p2)
        l2 = Line.of_end_points(p2, p3)
        l3 = Line.of_end_points(p3, p1)

        area = Area.of_lines((l1, l2, l3))

        # act
        # assert
        with self.assertRaises(IndexError) as context:
            area.get_line(-5)

        self.assertTrue('The requested line index -5 is out of range for a list of length 3' in str(context.exception))

    def test_rotate(self):
        # arrange
        p1 = Point.of_polar(10, 10)
        p2 = Point.of_polar(20, 30)
        p3 = Point.of_polar(15, 45)


        l1 = Line.of_end_points(p1, p2)
        l2 = Line.of_end_points(p2, p3)
        l3 = Line.of_end_points(p3, p1)

        alpha = 30

        area = Area.of_lines((l1, l2, l3))

        # act
        area_rotated = area.rotate(alpha)

        # assert
        # # line 1
        self.assertAlmostEqual(Point.of_polar(10, 40).x, area_rotated.get_line(0).p1.x, places=6)
        self.assertAlmostEqual(Point.of_polar(10, 40).y, area_rotated.get_line(0).p1.y, places=6)
        self.assertAlmostEqual(Point.of_polar(20, 60).x, area_rotated.get_line(0).p2.x, places=6)
        self.assertAlmostEqual(Point.of_polar(20, 60).y, area_rotated.get_line(0).p2.y, places=6)

        # # line 2
        self.assertAlmostEqual(Point.of_polar(20, 60).x, area_rotated.get_line(1).p1.x, places=6)
        self.assertAlmostEqual(Point.of_polar(20, 60).y, area_rotated.get_line(1).p1.y, places=6)
        self.assertAlmostEqual(Point.of_polar(15, 75).x, area_rotated.get_line(1).p2.x, places=6)
        self.assertAlmostEqual(Point.of_polar(15, 75).y, area_rotated.get_line(1).p2.y, places=6)

        # # line 3
        self.assertAlmostEqual(Point.of_polar(15, 75).x, area_rotated.get_line(2).p1.x, places=6)
        self.assertAlmostEqual(Point.of_polar(15, 75).y, area_rotated.get_line(2).p1.y, places=6)
        self.assertAlmostEqual(Point.of_polar(10, 40).x, area_rotated.get_line(2).p2.x, places=6)
        self.assertAlmostEqual(Point.of_polar(10, 40).y, area_rotated.get_line(2).p2.y, places=6)

    def test_str(self):
        # arrange
        p1 = Point.of_polar(10, 10)
        p2 = Point.of_polar(20, 30)
        p3 = Point.of_polar(15, 45)


        l1 = Line.of_end_points(p1, p2)
        l2 = Line.of_end_points(p2, p3)
        l3 = Line.of_end_points(p3, p1)

        alpha = 30

        area = Area.of_lines((l1, l2, l3))

        # act
        str_act = str(area)

        # assert
        self.assertEqual('(9.848078, 1.736482) -> (17.320508, 10.000000),(17.320508, 10.000000) -> '
                         '(10.606602, 10.606602),(10.606602, 10.606602) -> (9.848078, 1.736482)', str_act)

    def test_find_trapezoid_shift_to_intercept_with_radius_no_interception(self):
        # arrange
        p1 = Point.of_cartesian(10, 10)
        p2 = Point.of_cartesian(20, 10)
        p3 = Point.of_cartesian(20, 20)
        p4 = Point.of_cartesian(10, 20)

        l1 = Line.of_end_points(p1, p2)
        l2 = Line.of_end_points(p2, p3)
        l3 = Line.of_end_points(p3, p4)
        l4 = Line.of_end_points(p4, p1)

        area = Area.of_lines((l1, l2, l3, l4))

        # act
        shift = Area.find_trapezoid_shift_to_intercept_with_radius(1, area)

        # assert
        self.assertAlmostEqual(0.0, shift, places=1)

    def test_immutability(self):
        # arrange
        p1 = Point.of_polar(10, 10)
        p2 = Point.of_polar(20, 30)
        p3 = Point.of_polar(15, 45)


        l1 = Line.of_end_points(p1, p2)
        l2 = Line.of_end_points(p2, p3)
        l3 = Line.of_end_points(p3, p1)

        area = Area.of_lines((l1, l2, l3))

        # assert
        with self.assertRaises(dataclasses.FrozenInstanceError):
            exec('area.lines = (l3, l2, l1)')



from dataclasses import dataclass

from magnumapi.cadata.Definition import Definition


@dataclass
class FilamentDefinition(Definition):
    """Class for strand definition.

       Attributes:
           d_fil_out (float): The outer diameter of filament (um).
           d_fil_in (float): The inner diameter of filament (um).
           fit_j_c (str): Critical current fit.
           fit_perp (str): Name of the critical surface fit for orthogonal direction 2.
    """
    d_fil_out: float
    d_fil_in: float
    fit_j_c: str
    fit_perp: str

    @staticmethod
    def get_magnum_to_roxie_dct() -> dict:
        return {"name": "Name",
                "d_fil_out": "fildiao",
                "d_fil_in": "fildiai",
                "fit_j_c": "Jc-Fit",
                "fit_perp": "fit-|",
                "comment": "Comment"}



from abc import ABC, abstractmethod

import pandas as pd


class ToolAdapter(ABC):
    """An abstract class providing requirements for a ToolAdapter implementation

    """

    @abstractmethod
    def run(self) -> None:
        """ Abstract method running a tool adapter

        """
        raise NotImplementedError('This method is not implemented for this class')

    @staticmethod
    @abstractmethod
    def convert_figures_of_merit_to_dict(fom_df: pd.DataFrame) -> dict:
        """ Abstract method converting a dataframe with figures of merit into a dictionary

        :return: a dictionary mapping a figure of merit name to its value
        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def read_figures_of_merit_table(self) -> pd.DataFrame:
        """ Abstract method reading a dataframe with figures of merit

        :return: a dataframe with figures of merit
        """
        raise NotImplementedError('This method is not implemented for this class')


import json

def read(json_file_path: str) -> dict:
    """ Static method reading a json file and returning a list of dictionaries with block definitions.

    :param json_file_path: a path to a json file
    :return: a list of dictionaries with geometry definition (block definition)
    """
    #todo: add check if the file exists - as a decorator
    with open(json_file_path) as f:
        return json.load(f)